! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module convertStringToInteger_mod

use dal
use oal
use errorHandling

implicit none

private

public convertExposureStringToInteger

contains

!
!------------------------------------------------------------------------
!
! SUBROUTINE ConvertExposureStringToInteger
!
! Written by:
!                  Andrew King
! Date : 24/07/2000
!
! Purpose: To convert a string number representing the exposure number to an integer value.
!
! Parameters:
!            In : exposureString - length (in fact of length 3)
!
!            Out: exposureNumber - integer value of the exposure.
!
! Method:
! Process each character in the string starting at the the extreme right hand side. Convert each character
! to an integer and multiply by the appropriate power of ten.
!
!------------------------------------------------------------------------ 
subroutine convertExposureStringToInteger(exposureString ,exposureNumber)
  
  character(len=*),intent(in)           :: exposureString ! Input exposure number string.
  
  integer,intent(out)                   :: exposureNumber ! Output integer value of the exposureString.

  integer(kind=single)                  :: powerOfTen     ! Counter for power of 10.    
  integer(kind=single)                  :: nCounter       ! Loop counter. 
  integer(kind=single)                  :: nPos           ! Position in string.

!
! Initialise.
!
  powerOfTen = 0
  exposureNumber = 0
  nPos = len_trim(exposureString)

!
! Loop through the characters one-by-one
!
  do nCounter = 1,len_trim(exposureString)
!
! Obtain the next character, conver to an integer and multiply by appopriate power of then. Add in to accumulator.
! 
     exposureNumber = exposureNumber + (ichar(exposureString(nPos-nCounter+1:nPos-nCounter+1)) & 
          - ichar('0')) * 10**powerOfTen 

!
! Move up one character in the sring.
!
!     nPos = nPos - 1
!
! Increment power of ten
!
     powerOfTen = powerOfTen + 1
  end do

  return

end subroutine convertExposureStringToInteger

end module convertStringToInteger_mod










