! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module copyImage_mod

   use dal
   use sasdal
   use errorHandling

   implicit none

   private

   public copyImage

contains

subroutine copyImage(sourceDataSet, destDataSet, arrayName, arrayType)

   type(DataSetT) :: sourceDataSet, destDataSet
   type(ArrayT) :: sourceArray, destArray
   integer :: arrayType
   character(len=*) :: arrayName
   real(kind=single),pointer :: temp_r32(:,:), temp2_r32(:,:)
   integer(kind=Int32),pointer :: temp_i32(:,:), temp2_i32(:,:)
   integer(kind=Int8),pointer :: temp_i8(:,:), temp2_i8(:,:)

   sourceArray = array(sourceDataSet, arrayName, read)


   if (arrayType == Real32) then
      temp_r32 => real32Array2Data(sourceArray)
      destArray = addArray(destDataSet, arrayName, arrayType, shape(temp_r32))
      temp2_r32 => real32Array2Data(destArray)
      temp2_r32 = temp_r32 
   else if (arrayType == Integer32) then
      temp_i32 => int32Array2Data(sourceArray)
      destArray = addArray(destDataSet, arrayName, arrayType, shape(temp_i32))
      temp2_i32 => int32Array2Data(destArray)
      temp2_i32 = temp_i32 
   else if (arrayType == Integer8) then
      temp_i8 => int8Array2Data(sourceArray)
      destArray = addArray(destDataSet, arrayName, arrayType, shape(temp_i8))
      temp2_i8 => int8Array2Data(destArray)
      temp2_i8 = temp_i8
   else
      call fatal("copyImage","Cannot deal with array type")
   endif

   destArray = sourceArray

   call copyAllAttributes(attributable(sourceArray), attributable(destArray))
   call release(destArray)
   call release(sourceArray)

end subroutine copyImage

end module copyImage_mod
