! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module Matrix_mod

IMPLICIT NONE

PRIVATE
PUBLIC :: Ludcmp, Lubksb

CONTAINS

!*********************************************
! 
!********************************************
subroutine Ludcmp(n, a, index, d)
    integer, intent(in) :: n
    real, intent(inout) :: a(1 : n, 1 : n)
    integer, intent(out) :: index(1 : n)
    real, intent(out) :: d
    integer :: i, imax, j, k, l, m
    real :: big, dum, sum, temp
    real, allocatable :: vv(:)
    real, parameter :: tiny =  1.0e-20

    allocate(vv(1 : n))

    d = 1.0

    do i = 1, n
        big = 0.0
        do j = 1, n
            temp = abs(a(i, j))
            if (temp .gt. big) big = temp
        end do
        if (big .eq. 0.0) then
            write(*, *) "Singular matrix in subroutine ludcmp"
            do l = 1, n
                do m = 1, n
                    write(*, *) "a(", l, ",", m, ")=", a(l, m)
                end do
            end do 
            stop
        end if
        vv(i) = 1.0 / big
       
    end do

    do j = 1, n
        do i = 1, j - 1
            sum = a(i, j)
            do k = 1, i - 1
                sum = sum - a(i, k) * a(k, j)
            end do
            a(i, j) = sum
        end do
        big = 0.0
        do i = j, n
            sum = a(i, j)
            do k = 1, j - 1
                sum = sum - a(i, k) * a(k, j)
            end do
            a(i, j) = sum
            dum = vv(i) * abs(sum)
            if (dum .ge. big) then
                big = dum
                imax = i
            end if
        end do
        if (j .ne. imax) then
            do k = 1, n
                dum = a(imax, k)
                a(imax, k) = a(j, k)
                a(j, k) = dum
            end do
            d = -d
            vv(imax) = vv(j)
        end if 
        index(j) = imax
        if (a(j, j) .eq. 0.0) a(j, j) = tiny    
         
        if (j .ne. n) then
            dum = 1.0 / a(j, j)
            do i = j + 1, n
                a(i, j) = a(i, j) * dum
            end do
        end if
    end do
    deallocate(vv)
end subroutine Ludcmp

!*********************************************
! Solves the set of n linear equations A.X=B
! a is input as its LU decomposition
! 
!********************************************
subroutine Lubksb(n, a, index, b)
    integer, intent(in) :: n
    real, intent(in) :: a(1 : n, 1 : n)
    integer, intent(in) :: index(1 : n)
    real, intent(inout) :: b(1 : n)

    integer :: i, ii, ip, j

    real :: sum
    
    ii = 0

    do i = 1, n
        ip = index(i)
        sum = b(ip)
        b(ip) = b(i)
        if (ii .gt. 0) then
            do j = ii, i - 1
                sum = sum - a(i, j) * b(j)
            end do
        else if (sum .gt. 0.0) then 
            ii = i
        end if
        b(i) = sum
    end do

    do i = n, 1, -1
        sum = b(i)
        do j = i + 1, n
            sum = sum - a(i, j) * b(j)
        end do
        b(i) = sum / a(i, i)
    end do

end subroutine Lubksb     
                
end module Matrix_mod 




