! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module omellipse_mod
   use types
   use omconstants_mod

contains

!*********************************************************
!
!*********************************************************
real(single) Function PixelArea(x, y, xc, yc, a, b, theta1)
    real(single), intent(in) :: x, y, xc, yc, a, b, theta1
    real(single) :: x1, y1, x2, y2
    real(single) :: theta, number
    integer(int8) :: count
    integer(int32) :: i, n, n1

    theta = theta1 * degreesToRadians ! NB Need theta in radians
    x1 = x - 0.5  ! left x vertex of pixel
    y1 = y - 0.5  ! lower y vertex of pixel
    x2 = x + 0.5  ! right x vertex of pixel
    y2 = y + 0.5  ! upper y vertex of pixel



    count = 0
    
    ! Distance from xc, yc to lower left vertex
    
    if (PointInEllipse(x1, y1, xc, yc, a, b, theta)) count = count + 1

    if (PointInEllipse(x2, y1, xc, yc, a, b, theta)) count = count + 1


    if (PointInEllipse(x2, y2, xc, yc, a, b, theta)) count = count + 1

   
    if (PointInEllipse(x1, y2, xc, yc, a, b, theta)) count = count + 1

    if (count == 4) then
        PixelArea = 1.0
        return
    end if
 
    if (count == 0) then
        PixelArea = 0.0
        return
    end if
 

    n = 10000
    n1 = 0
    do i = 1, n
        call random_number(number)
        x2 = x1 + number
        call random_number(number)
   
        y2 = y1 + number
      
        if (PointInEllipse(x2, y2, xc, yc, a, b, theta)) n1 = n1 + 1
             
    end do
    PixelArea = real(n1) / real(n)
end function PixelArea
!********************************************************
!
!********************************************************
logical Function PointInEllipse(xPos, yPos, xc, yc, a, b, theta)
    real(single), intent(in) :: xPos, yPos, xc, yc, a, b, theta
    real(single) :: dx, dy, x, y, res

    dx = xPos - xc
    dy = yPos - yc
    call RectangToEllipse(dx, dy, theta, x, y)
    res = x * x / (a * a) + y * y / (b * b)

    if (res <= 1.0) then
        PointInEllipse = .true.
    else
        PointInEllipse = .false.
    end if
end function PointInEllipse
!******************************************************
!
!*****************************************************
subroutine RectangToEllipse(xIn, yIn, theta, xOut, yOut)
    real(single), intent(in) :: xIn, yIn, theta
    real(single), intent(out) :: xOut, yOut
    real(single) :: sinTheta, cosTheta

    sinTheta = sin(theta)
    cosTheta = cos(theta)

    

!    xOut = xIn * cos(theta) - yIn * sin(theta)
!    yOut = xIn * sin(theta) + yIn * cos(theta)

    xOut = xIn * costheta + yIn * sintheta
    yOut = yIn * costheta - xin * sintheta


end subroutine RectangToEllipse

end module omellipse_mod





