! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module omgraph_mod
    use param
    use dal
    use errorHandling
 

    use omallocation_mod

interface DrawGraph
    module procedure DrawGraph1, DrawGraph2
end interface DrawGraph



interface Histogram
    module procedure Histogram1, Histogram2, &
     Histogram3, Histogram4, Histogram5
end interface Histogram
real(single), private :: dy = 2.0

contains
!****************************************************************
!
!
!****************************************************************

subroutine DrawGraph1(ndata, xdata, ydata, &
         title, xlabel, ylabel, x1, x2, y1, y2, &
         xmin1, xmax1, ymin1, ymax1)
  
    integer, intent(in) :: ndata
    real, intent(in) :: xdata(1 : ndata)
    real, intent(in) :: ydata(1 : ndata)
   

    character(len=*), intent(in) :: title
    character(len=*), intent(in) :: xlabel
    character(len=*), intent(in) :: ylabel
   
    real, intent(in) :: x1, x2, y1, y2
    real, intent(in), optional :: xmin1, xmax1, ymin1, ymax1


    real ::  xmin, xmax, ymin, ymax
    integer j
       
    xmin =  10000.0
    xmax = -10000.0
    ymin = xmin
    ymax = xmax
             
    do j = 1, ndata
        
        if (xdata(j) .lt. xmin) xmin = xdata(j)
        if (xdata(j) .gt. xmax) xmax = xdata(j)
        if (ydata(j) .lt. ymin) ymin = ydata(j)
        if (ydata(j) .gt. ymax) ymax = ydata(j)
    end do
    if (abs(ymax - ymin) .lt. 1e-10) then
        ymin = ymin - 0.1
        ymax = ymax + 0.1
    end if    

    
    call pgsvp(x1, x2, y1, y2)

     
    call pgrnge(xmin, xmax, xmin, xmax)
    call pgrnge(ymin, ymax, ymin, ymax)
    if (present(xmin1)) xmin = xmin1
    if (present(xmax1)) xmax = xmax1
    if (present(ymin1)) ymin = ymin1
    if (present(ymax1)) ymax = ymax1




    call pgswin(xmin, xmax, ymin, ymax)
    call pgsch(1.0)
  
  
    call pgslw(1)
    call pgbox("bctsi", 0.0, 0, "bctsi", 0.0, 0)
    call pgsch(0.8)
    call pgslw(1)
    call pgbox("n", 0.0, 0, "n", 0.0, 0)

    call pgsch(1.0) 
    call pgslw(1)
    call pglabel("", ylabel, title)

    call pgmtxt('B', dy, 0.5, 0.5, trim(xlabel))
        !





    
    call pgslw(1)
  
    call pgsch(0.6)        
    call pgpoint(ndata, xdata, ydata, 9)
   

end subroutine DrawGraph1

!****************************************************************
!
!
!****************************************************************
subroutine DrawGraph2(ndata, xdata, xerrors, ydata, yerrors, &
         title, xlabel, ylabel, x1, x2, y1, y2, &
         xmin1, xmax1, ymin1, ymax1, equalaxes)
  
    integer, intent(in) :: ndata
    real, intent(in) :: xdata(1 : ndata)
    real, intent(in) :: xerrors(1 : ndata)
    real, intent(in) :: ydata(1 : ndata)
    real, intent(in) :: yerrors(1 : ndata)

    character(len=*), intent(in) :: title
    character(len=*), intent(in) :: xlabel
    character(len=*), intent(in) :: ylabel
   
    real, intent(in) :: x1, x2, y1, y2
    real, intent(in), optional :: xmin1, xmax1, ymin1, ymax1
    logical, intent(in), optional :: equalAxes

    real, dimension( : ), allocatable :: lower, upper
    real ::  xmin, xmax, ymin, ymax
    integer j
       
    xmin =  10000.0
    xmax = -10000.0
    ymin = xmin
    ymax = xmax
             
    do j = 1, ndata
        
        if (xdata(j) .lt. xmin) xmin = xdata(j)
        if (xdata(j) .gt. xmax) xmax = xdata(j)
        if (ydata(j) .lt. ymin) ymin = ydata(j)
        if (ydata(j) .gt. ymax) ymax = ydata(j)
    end do
    if (abs(ymax - ymin) .lt. 1e-10) then
        ymin = ymin - 0.1
        ymax = ymax + 0.1
    end if    

    
    call pgsvp(x1, x2, y1, y2)

     
    call pgrnge(xmin, xmax, xmin, xmax)
    call pgrnge(ymin, ymax, ymin, ymax)
    if (present(xmin1)) xmin = xmin1
    if (present(xmax1)) xmax = xmax1
    if (present(ymin1)) ymin = ymin1
    if (present(ymax1)) ymax = ymax1



     if ( present(equalAxes)) then
         if (equalAxes) then 
             call pgwnad(xmin, xmax, ymin, ymax)
         else
             call pgswin(xmin, xmax, ymin, ymax)
         end if
     else
       call pgswin(xmin, xmax, ymin, ymax)
     end if

  call pgsch(1.0)
  
  
    call pgslw(1)
    call pgbox("bctsi", 0.0, 0, "bctsi", 0.0, 0)
    call pgsch(0.8)
    call pgslw(1)
    call pgbox("n", 0.0, 0, "n", 0.0, 0)

    call pgsch(1.0) 
    call pgslw(1)
    call pglabel("", ylabel, title)
    call pgmtxt('B', dy, 0.5, 0.5, trim(xlabel))


    call pgslw(1)
  
    call pgsch(0.6)        
    call pgpoint(ndata, xdata, ydata, 9)
   
!    call pgline(ndata, xdata, ydata)

       
    allocate(lower(1 : ndata))
    allocate(upper(1 : ndata))
     
    lower = xdata - xerrors
    upper = xdata + xerrors
           
    call pgerrx(ndata, lower, upper, ydata, 0)
          
    lower = ydata - yerrors
    upper = ydata + yerrors
           
    call pgerry(ndata, xdata, lower, upper, 0)
    deallocate(lower)
    deallocate(upper)  

    call pgslw(1)
    call pgsch(0.5)
 
end subroutine DrawGraph2


!*****************************************************
! Subroutine Histogram
!
!*****************************************************
subroutine Histogram1(ndata, data, noBins, &
                 xLabel, title, x1, x2, y1, y2)
    integer, intent(in) :: ndata
    real(single), intent(in), dimension(ndata) :: data
    integer(int32), intent(in) :: noBins
    character(*), intent(in) :: title, xLabel
    real, intent(in) :: x1, x2, y1, y2
    real(single) ::  xmin, xmax, charSize, yMax, binWidth     
    integer i
    integer(int32), pointer:: frequency(:) 
    
    integer :: k = 1
    character(100) :: yLabel 

    xmin = minval(data)
    xmax = maxval(data)


    if (abs(xmin - xmax) .lt. 1e-6) then
        call warning("Unable to draw histogram", "xmin=xmax")
        return
    end if

    charSize = (x2 - x1) * 1.5

    xmin = 1e6
    xmax = -1e6
    do i = 1, ndata
        xmin = min(xmin, data(i))
        xmax = max(xmax, data(i))
    end do


    

    !******************************
    ! Find the maximum bin size
    !******************************

    frequency => AllocateInt32(1, noBins, "frequency")
    frequency = 0
  
    do i = 1, ndata
        k = int((data(i) - xmin) / binWidth + 0.00000001) + 1       
        k = max(1, k)
        k = min(noBins, k)
        frequency(k) = frequency(k) + 1
        
    end do
    
    yMax = int(real(maxval(frequency)) * 1.50) 

    deallocate(frequency)  
     
    call pgsvp(x1, x2, y1, y2)

    call pgwindow(xmin, xmax, 0.0, yMax)
   
     
    call pgsch(1.0) 
    call pgsfs(2)    
    call pgslw(1)    
    call pghist(ndata, data, xmin, xmax, noBins, 3)

    call pgwindow(xmin, xmax, 0.0, (yMax / real(ndata)) * 100.0)
  

    
    
    call pgbox("bctsi", binWidth, 5, "bctsi", 0.0, 0)

    call pgsch(0.9)  
    call pgbox("n", binWidth, 5, "n", 0.0, 0) 
  

    call pgsch(1.0)
    write(yLabel, "(a, i5, a)") "FREQUENCY (%) (TOTAL=", ndata, ")"
       
    call pglabel("", yLabel, title)
       call pgmtxt('B', dy, 0.5, 0.5, trim(xlabel))


end subroutine Histogram1


!*****************************************************
! Subroutine Histogram
!
!*****************************************************
subroutine Histogram2(ndata, data, binWidth, &
                 xLabel, title, x1, x2, y1, y2)
    integer, intent(in) :: ndata
    real(single), intent(in), dimension(ndata) :: data
    real(single), intent(inout) :: binWidth
    character(*), intent(in) :: title, xLabel
    real, intent(in) :: x1, x2, y1, y2
    real(single) :: xmin, xmax, ymax, charSize        
    integer i, j, noBins
    integer(int32), pointer:: frequency(:) 
  
    integer :: k = 1
    character(100) :: yLabel 

    xmin = minval(data)
    xmax = maxval(data)


    if (abs(xmin - xmax) .lt. 1e-6) then
        call warning("Unable to draw histogram", "xmin=xmax")
        return
    end if

    charSize = (x2 - x1) * 1.5

    xmin = 1e6
    xmax = -1e6
    do i = 1, ndata
        xmin = min(xmin, data(i))
        xmax = max(xmax, data(i))
    end do


    xmin = real(int(xmin))

    do j = 1, 10
        noBins = int(ceiling((real(ceiling(xmax)) - xmin) / binWidth))
        if (noBins < 16 .and. noBins > 5) exit
        if (noBins < 5) then
            binWidth = binWidth / 2.0
        else
           binWidth = binWidth * 2.0
        end if 
    end do
    if (noBins < 3) then
        call warning("omgraph", "No of bins < 3")
        return
    end if
    xmax = xmin + real(noBins) * binWidth   
    

    !******************************
    ! Find the maximum bin size
    !******************************

    frequency => AllocateInt32(1, noBins, "frequency")
    frequency = 0
    do i = 1, ndata
        k = int((data(i) - xmin) / binWidth + 0.00000001) + 1       
        k = max(1, k)
        k = min(noBins, k)
        frequency(k) = frequency(k) + 1
       
    end do
    
    yMax = int(real(maxval(frequency)) * 1.50) 

    deallocate(frequency)  
     
    call pgsvp(x1, x2, y1, y2)

    call pgwindow(xmin, xmax, 0.0, yMax)
   
     
    call pgsch(1.0) 
    call pgsfs(2)    
    call pgslw(1)    
    call pghist(ndata, data, xmin, xmax, noBins, 3)

    call pgwindow(xmin, xmax, 0.0, (yMax / real(ndata)) * 100.0)
  

    call pgsch(0.8)  
 
    
    call pgbox("bctsi", binWidth, 5, "bctsi", 0.0, 0)

    call pgsch(0.8)  
    call pgbox("n", binWidth, 5, "n", 0.0, 0) 
  

    call pgsch(1.0)
    write(yLabel, "(a, i5, a)") "FREQUENCY (%) (TOTAL=", ndata, ")"
       
    call pglabel("", yLabel, title)
    
   call pgmtxt('B', dy, 0.5, 0.5, trim(xlabel))

end subroutine Histogram2


!*****************************************************
! Subroutine Histogram
!
!*****************************************************
subroutine Histogram3(ndata, data, noBins, &
                 xLabel, title, x1, x2, y1, y2, &
                 xmin, xmax)
    integer, intent(in) :: ndata
    real(single), intent(in), dimension(ndata) :: data
    integer(int32), intent(in) :: noBins
    character(*), intent(in) :: title, xLabel
    real, intent(in) :: x1, x2, y1, y2
    real(single), intent(in) ::  xmin, xmax
    real(single) :: charSize, yMax, binWidth     
    integer i
    integer(int32), pointer:: frequency(:) 
    
    integer :: k = 1
    character(100) :: yLabel 

 


    if (abs(xmin - xmax) .lt. 1e-6) then
        call warning("Unable to draw histogram", "xmin=xmax")
        return
    end if

    charSize = (x2 - x1) * 1.5

   
    

    !******************************
    ! Find the maximum bin size
    !******************************

    frequency => AllocateInt32(1, noBins, "frequency")
    frequency = 0
  
    do i = 1, ndata
        k = int((data(i) - xmin) / binWidth + 0.00000001) + 1       
        k = max(1, k)
        k = min(noBins, k)
        frequency(k) = frequency(k) + 1
        
    end do
    
    yMax = int(real(maxval(frequency)) * 1.50) 

    deallocate(frequency)  
     
    call pgsvp(x1, x2, y1, y2)

    call pgwindow(xmin, xmax, 0.0, yMax)
   
     
    call pgsch(0.8) 
    
    call pgslw(1)    
    call pghist(ndata, data, xmin, xmax, noBins, 3)

    call pgwindow(xmin, xmax, 0.0, (yMax / real(ndata)) * 100.0)
  

    
    call pgsch(0.8)  

    call pgbox("bctsi", binWidth, 5, "bctsi", 0.0, 0)

    call pgsch(1.0)  
    call pgbox("n", binWidth, 5, "n", 0.0, 0) 
  

    call pgsch(1.0)
    write(yLabel, "(a, i5, a)") "FREQUENCY (%) (TOTAL=", ndata, ")"
       
    call pglabel("", yLabel, title)
    
   call pgmtxt('B', dy, 0.5, 0.5, trim(xlabel))

end subroutine Histogram3


!*****************************************************
! Subroutine Histogram
!
!*****************************************************
subroutine Histogram4(ndata, data, binWidth, &
                 xLabel, title, x1, x2, y1, y2, &
                 xmin, xmax)
    integer, intent(in) :: ndata
    real(single), intent(in), dimension(ndata) :: data
    real(single), intent(inout) :: binWidth
    character(*), intent(in) :: title, xLabel
    real, intent(in) :: x1, x2, y1, y2
    real(single), intent(in) ::  xmin, xmax
    real(single) :: charSize, yMax     
         
    integer i, noBins
    integer(int32), pointer:: frequency(:) 
  
    integer :: k = 1
    character(len=100) :: yLabel 

    


    if (abs(xmin - xmax) .lt. 1e-6) then
        call warning("Unable to draw histogram", "xmin=xmax")
        return
    end if

    charSize = (x2 - x1) * 1.5

   

    
    noBins = int(ceiling((real(ceiling(xmax)) - xmin) / binWidth))
    if (noBins < 3) then
        call warning("omgraph", "No of bins < 3")
        return
    end if
            
    

    !******************************
    ! Find the maximum bin size
    !******************************

    frequency => AllocateInt32(1, noBins, "frequency")
    frequency = 0
    do i = 1, ndata
        k = int((data(i) - xmin) / binWidth + 0.00000001) + 1       
        k = max(1, k)
        k = min(noBins, k)
        frequency(k) = frequency(k) + 1
       
    end do
    
    yMax = int(real(maxval(frequency)) * 1.50) 

    deallocate(frequency)  
     
    call pgsvp(x1, x2, y1, y2)

    call pgwindow(xmin, xmax, 0.0, yMax)
   
     
    call pgsch(0.8) 
    call pgsfs(2)    
    call pgslw(1)    
    call pghist(ndata, data, xmin, xmax, noBins, 3)

    call pgwindow(xmin, xmax, 0.0, (yMax / real(ndata)) * 100.0)
  

    
    
    call pgbox("bctsi", binWidth, 5, "bctsi", 0.0, 0)

    call pgsch(0.8)  
    call pgbox("n", binWidth, 5, "n", 0.0, 0) 
  

    call pgsch(1.0)
    write(yLabel, "(a, i5, a)") "FREQUENCY (%) (TOTAL=", ndata, ")"
       
    call pglabel("", yLabel, title)
    call pgmtxt('B', dy, 0.5, 0.5, trim(xlabel))


end subroutine Histogram4








!*****************************************************
! Subroutine Histogram
!
!*****************************************************
subroutine Histogram5(nHists, ndata, data, binWidth, &
                 xLabel, title, caption, x1, x2, y1, y2, &
                 xmin, xmax)
    integer(int32), intent(in) :: nHists
    integer(int32), intent(in) :: ndata(:)
    real(single), intent(in), dimension(:, :) :: data
    real(single), intent(inout) :: binWidth
    character(*), intent(in) :: title, xLabel
    character(*), intent(in), optional :: caption(:)

    real, intent(in) :: x1, x2, y1, y2
    real(single), intent(in) ::  xmin, xmax
    real(single) :: charSize, yMax     
    real(single), pointer :: tempData(:) => null()         
    integer i, j, noBins
    integer(int32), pointer:: frequency(:) 
  
    integer :: k = 1
    character(100) :: yLabel, text 


    if (abs(xmin - xmax) .lt. 1e-6) then
        call warning("Unable to draw histogram", "xmin=xmax")
        return
    end if

    charSize = (x2 - x1) * 1.5

   

    
    noBins = int(ceiling((real(ceiling(xmax)) - xmin) / binWidth))
    if (noBins < 3) then
        call warning("omgraph", "No of bins < 3")
        return
    end if
            
    

    !******************************
    ! Find the maximum bin size
    !******************************

    frequency => AllocateInt32(1, noBins, "frequency")
    frequency = 0
    yMax = 0.0
    do j = 1, nHists    
        frequency = 0
        
        do i = 1, ndata(j)
            k = int((data(j, i) - xmin) / binWidth + 0.00000001) + 1       
            k = max(1, k)
            k = min(noBins, k)
            frequency(k) = frequency(k) + 1
        end do
        yMax = max(yMax, real(int(real(maxval(frequency)) * 1.50))) 

    end do
    
  
    deallocate(frequency)  
     
    call pgsvp(x1, x2, y1, y2)

    call pgwindow(xmin, xmax, 0.0, yMax)
   
     
    call pgsch(1.0) 
    call pgsfs(2)    
    call pgslw(1)    

    do i = 1, nHists
        tempData => allocateSingle(1, ndata(i), "tempData")
        tempData(1 : nData(i)) = data(i,  1 : nData(i))
        call pgsci(1)
        call pgsfs(2)

        call pghist(ndata(i), tempData, xmin, xmax, noBins, 3)
        if (i > 1) then
            call pgsci(i)
            call pgsfs(i + 1)

            call pghist(ndata(i), tempData, xmin, xmax, noBins, 3)
        end if

        deallocate(tempData)
        if (present(caption)) then
            call pgsch(0.8)
 
            call pgmtxt('T', real(-i), 0.0, 0.0, trim(caption(i)))
   
        end if
    end do

    call pgsci(1)


!    call pgwindow(xmin, xmax, 0.0, (yMax / real(ndata(1))) * 100.0)
  

    
    call pgsch(0.8)
    call pgbox("bctsi", binWidth, 5, "bctsi", 0.0, 0)

    call pgsch(0.8)  
    call pgbox("n", binWidth, 5, "n", 0.0, 0) 
  

    call pgsch(0.9)
!    write(text, "(a)") "NO", ndata(1), ")"
       
    call pglabel("", "NUMBER", title)

      call pgmtxt('B', dy, 0.5, 0.5, trim(xlabel))
text = "trext"
ylabel="ylabel"
        write(*, *) "text=", text, " xlabel=", ylabel

end subroutine Histogram5


end module omgraph_mod
