! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

!********************************************
! module OmdetectPhotometryDefs_mod
! For use with module omdetectAperturePhotometry
!
!********************************************
module OmPhotometryDefs_mod
    use dal

    type OMpoint
        real(single) :: x
        real(single) :: y
        integer(int8) :: extension
        real(single) :: maxCounts
    end type OMPOINT


    integer(int16) :: sourceNumber    ! The number of the source 

    integer :: noPointSources  

    !***********************************************************
    ! Array CCDImage will store the 2-d background image
    !***********************************************************

    integer(int16) :: nx, ny
    real(single), dimension(: , :), pointer :: CCDImage => null()
   

    ! Smoothed image to be used to get the background
    real(single), dimension(: , :), pointer :: backGroundImage => null()



    ! To store a possible exposure-map image
    real(single), dimension(: , :), pointer :: exposureImage => null()


    integer(int16), dimension(:, :), pointer :: sourceNumImage => null()



  
    logical(boolean), dimension(: , :), pointer :: qualityImage => null()



    integer(int32), dimension(:), pointer :: sortIndex => null()
    type(OMPOINT), dimension(:), pointer :: sourceCoords => null()
 
    !***********************************************************
    ! The following will start the array dimensions of the images
    !***********************************************************
    integer(int32) :: il = 0  ! The start x index
    integer(int32) :: jl = 0  ! The start y index
    integer(int32) :: iu      ! The finish x index
    integer(int32) :: ju      ! The finish y index
    real(single) :: xLeft
    real(single):: yBottom
    real(single):: xRight
    real(single):: yTop

    real(single) :: apertureRadius
    real(single)  :: coincLossRadius     

    real(double)  :: apertureArea   ! The actual aperture area used to get counts
    real(single)  :: fullApertureArea   ! The required aperture area used to get counts
    
 

   
    !************************************************************************
    ! The following will store the maximum of the x amd y fwhm of each object
    !*************************************************************************
    real (single), dimension(:), pointer :: objectWidth => null()
 
   
 
    !************************************************************************
    ! The following are used to compute the median background around a source
    !************************************************************************

    real (single), dimension(:), pointer :: backgroundPixel => null()
   

    real(double) :: backgroundCountsMode
    real(double) :: backgroundCountsModeError
    
    real (single) :: minimumApertureRadius    ! The minimum allowd aperture radius for doing photometry

    real (single) :: innerRadius              ! The inner radius for the background counts
  
   
    real (single) :: outerRadius              ! The outer radius for the background counts
    real (single) :: imageMode                ! The background mode value 
    integer(int32) :: maxNumPixels



    real(double) :: frameTime                 ! The frame time in seconds
    
    real(double) :: deadFrac

    real(single) :: exposureLength

 

end module OmPhotometryDefs_mod












