! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module OmUtility_mod

    use param
    use cal
    use dal
    use omallocation_mod
    use omsourceparam_mod
  
    use sla




!***********************************************************************
!
! The region colours
!**********************************************************************

character(7), dimension(0 : 11), private :: colours = (/"red    ",&
                                                 "magenta",      &
                                                 "yellow ",      &
                                                 "white  ",      &
                                                 "black  ",      &
                                                 "red    ",      &
                                                 "black  ",      & 
                                                 "blue   ",      &
                                                 "blue   ",      &
                                                 "red    ",      &
                                                 "red    ",      &
                                                 "red    "/)


public    

!****************************************************
!
!****************************************************
interface SaveImage
module procedure SaveRealImage1, SaveRealImage2, SaveInt8Image, SaveInt16Image



end interface SaveImage
   

interface AddColumnToTable
module procedure  AddReal32Column, AddReal64Column, AddInt8Column, &
                  AddInt16Column, AddInt32Column

end interface AddColumnToTable
interface ReadTableColumn
module procedure  ReadReal32Column, ReadReal64Column, ReadInt8Column, &
                  ReadInt16Column, ReadInt32Column, &
                  Read1Real32Column, Read1Real64Column, Read1Int8Column, &
                  Read1Int16Column, Read1Int32Column


end interface ReadTableColumn

interface ReadCharColumn
module procedure  ReadCharColumn, ReadCharColumn1


end interface ReadCharColumn



interface LoadVectorData

module procedure LoadRealVectorColumn,  loadDoubleVectorColumn, &
                 Loadint8VectorColumn,  loadint16VectorColumn, &
                 Loadint32VectorColumn

end interface LoadVectorData

interface AddVectorColumn
    module procedure AddRealVectorColumn1,      &
                     AddDoubleVectorColumn1,    &
                     AddInt8VectorColumn1,      &
                     AddInt16VectorColumn1,     &
                     AddInt32VectorColumn1



end interface addVectorColumn





interface isnullvalue

    module procedure isnullvalue1, isnullvalue2

end interface isnullvalue


   
   contains





!*******************************************************
! ModifiedJulianDate - Returns the modified Julian date
! from the string date representation
!*******************************************************
real(double) function ModifiedJulianDate(dateString)
    character(*), intent(in) :: dateString

    integer :: year            ! Stores the year from the date string
    integer :: month           ! Stores the month from the date string
    integer :: day             ! Stores the day from the date string  
    integer :: hour            ! Stores the hour from the date string
    integer :: minute          ! Stores the minute from the date string  
    integer :: second          ! Stores the second from the date string
    character(len = 1) char    ! Used for date and time extraction 
    ! from string representation
    real(double) JulianDay    
    integer :: flag

    read(dateString, '(i4, a1, i2, a1, i2, a1, i2, a1, i2, a1, i2)') &
         year, char, month, char, day, char, hour, &
         char, minute, char, second     

    ! Get the modified Julian date for this date
    call sla_CLDJ(year, month, day, JulianDay, flag)

    ! Need to add on the fraction of a day corresponding to the time
    ModifiedJulianDate = JulianDay + (dble(hour) + &
         (dble(minute) + dble(second) / 60.0d0) / 60.0d0) / 24.0d0
end function ModifiedJulianDate





!********************************************************************
! Attempts to read a string attribute from a fits file
! (Added by CB on 14.09.2000) because apparently some attributes may
! not be present.!
!********************************************************************
subroutine GetStringAttribute(dataSet, attribute, string)
    type(DataSetT), intent(in) :: dataSet
    character(*), intent(in) :: attribute
    character(*), intent(inout) :: string

    if (.not. hasAttribute(dataSet, attribute)) then
        string = ""
    else
        string = stringAttribute(dataSet, attribute)
    end if

end subroutine GetStringAttribute
!*********************************************************************
! Attempts to read a real attribute from a fits file
! (Added by CB on 14.09.2000) because apparently some attributes may
! not be present.
!*********************************************************************
Function GetReal32Attribute(dataSet, attribute, defaultValue) result(number)
    type(DataSetT), intent(in) :: dataSet
    character(*), intent(in) :: attribute
    real(single), intent(in), optional :: defaultValue
    real(single) :: number

    if (.not. hasAttribute(dataSet, attribute)) then
        if (present(defaultValue)) then
            number= defaultValue
        else
            number = 0.0
        end if
    else
        number = real32Attribute(dataSet, attribute)
    end if

end Function GetReal32Attribute


!*********************************************************************
! Attempts to read a real attribute from a fits file
! (Added by CB on 14.09.2000) because apparently some attributes may
! not be present.
!*********************************************************************
Function GetReal64Attribute(dataSet, attribute, defaultValue) result(number)
    type(DataSetT), intent(in) :: dataSet
    character(*), intent(in) :: attribute
    real(double), intent(in), optional :: defaultValue
    real(double) :: number

    if (.not. hasAttribute(dataSet, attribute)) then
        if (present(defaultValue)) then
            number= defaultValue
        else
            number = 0.0
        end if
    else
        number = real64Attribute(dataSet, attribute)
    end if

end Function GetReal64Attribute

!***************************************************
! AddReal32Cololumn
!***************************************************
subroutine AddReal32Column(tableSet, array, size, heading, &
        units, comment, nullValue, position)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading, units, comment
    integer, intent(in) :: size
    real(single), intent(in), dimension(:) :: array
    real(single), intent(in) :: nullValue
    integer(int32), intent(in), optional :: position
    real(single), pointer :: real32Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)
!write(*, *) "addreal32column - size=", size, " lbound array=", lbound(array), " ubound=", ubound(array)

    if (.not. hasColumn(tableSet, trim(heading))) then
        if (present(position)) then
            pColumn = addColumn(tableSet, trim(heading), real32, trim(units), ndims, trim(comment), position)
        else 
            pColumn = addColumn(tableSet, trim(heading), real32, trim(units), ndims, trim(comment))
        end if
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
   



    real32Col => real32Data(pcolumn)

    do i = 1 , size

        if (array(i) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            real32Col(i - 1) = array(i)
        end if
  
    enddo

    call release(pcolumn)

end subroutine AddReal32Column


!***************************************************
! AddReal64Column
!***************************************************


subroutine AddReal64Column(tableSet, array, size, heading, units, comment, &
        nullValue, position)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading, units, comment
    integer, intent(in) :: size

    real(double), intent(in), dimension(:) :: array
    real(single), intent(in) :: nullValue
    integer(int32), intent(in), optional :: position
    real(double), pointer :: real64Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)



    if (.not. hasColumn(tableSet, trim(heading))) then
        if (present(position)) then
            pColumn = addColumn(tableSet, trim(heading), real64, trim(units), ndims, trim(comment), position)
        else 
            pColumn = addColumn(tableSet, trim(heading), real64, trim(units), ndims, trim(comment))
        end if
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
   
   

    real64Col => real64Data(pcolumn)

    do i = 1 , size
        if (array(i) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            real64Col(i - 1) = array(i)
        end if
  
    enddo
    call release(pcolumn)

end subroutine AddReal64Column








!***************************************************
! AddInt8Cololumn
!***************************************************
subroutine AddInt8Column(tableSet, array, size, heading, &
           units, comment, nullValue, position)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading, units, comment
    integer, intent(in) :: size
    integer, intent(in), optional :: nullValue
    integer, intent(in), optional :: position  
    integer(int8), intent(in), dimension(:) :: array
    integer(int8), pointer :: int8Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)

    if (.not. hasColumn(tableSet, trim(heading))) then
        if(present(position)) then
            pColumn = addColumn(tableSet, trim(heading), integer8, &
                           trim(units), ndims, trim(comment), position)
        
        else 
            pColumn = addColumn(tableSet, trim(heading), integer8, trim(units), ndims, trim(comment))
        end if
    else
        pColumn = column(tableSet, trim(heading), modify)
 
    end if
   

    int8Col => int8Data(pcolumn)


    if (present(nullValue)) then
        call setNullValue(pcolumn, nullValue )
 
        do i = 1 , size
            if (array(i) .eq. nullValue) then
                call setToNull(pcolumn, i - 1)
            else
                int8Col(i - 1) = array(i)
            end if
  
        end do
    else
        forall(i = 1 : size)  int8Col(i - 1) = array(i)
    end if
 
    call release(pcolumn)

end subroutine AddInt8Column

!***************************************************
! AddInt16Cololumn
!***************************************************
subroutine AddInt16Column(tableSet, array, size, heading, &
           units, comment, nullValue, position)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading, units, comment
    integer, intent(in) :: size
    integer, intent(in), optional :: nullValue
    integer, intent(in), optional :: position  
    integer(int16), intent(in), dimension(:) :: array
    integer(int16), pointer :: int16Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)

    if (.not. hasColumn(tableSet, trim(heading))) then
        if(present(position)) then
            pColumn = addColumn(tableSet, trim(heading), integer16, &
                           trim(units), ndims, trim(comment), position)
        
        else 
            pColumn = addColumn(tableSet, trim(heading), integer16, trim(units), ndims, trim(comment))
        end if
    else
        pColumn = column(tableSet, trim(heading), modify)
 
    end if
   

    int16Col => int16Data(pcolumn)


    if (present(nullValue)) then
        call setNullValue(pcolumn, nullValue )
 
        do i = 1 , size
            if (array(i) .eq. nullValue) then
                call setToNull(pcolumn, i - 1)
            else
                int16Col(i - 1) = array(i)
            end if
  
        end do
    else
        forall(i = 1 : size)  int16Col(i - 1) = array(i)
    end if
 
    call release(pcolumn)

end subroutine AddInt16Column

!***************************************************
! AddInt32Cololumn
!***************************************************
subroutine AddInt32Column(tableSet, array, size, heading, units, &
           comment, nullValue)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading, units, comment
    integer, intent(in) :: size
    integer, intent(in), optional :: nullValue
    integer(int32), intent(in), dimension(:) :: array
    integer(int32), pointer :: int32Col(:)
    type(ColumnT) :: pColumn
    integer(int32) ::  i
    integer :: ndims(1) = (/1/)

    if (.not. hasColumn(tableSet, trim(heading))) then
        pColumn = addColumn(tableSet, trim(heading), integer32, trim(units), ndims, trim(comment))
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
   
   
    int32Col => int32Data(pcolumn)

    if (present(nullValue)) then
        call setNullValue(pcolumn, nullValue )
         
        do i = 1 , size
            if (array(i) .eq. nullValue) then
                call setToNull(pcolumn, i - 1)
            else
                int32Col(i - 1) = array(i)
            end if
  
        end do
    else
        forall(i = 1 : size)  int32Col(i - 1) = array(i)
    end if


    call release(pcolumn)

end subroutine AddInt32Column


!**********************************************************
! ReadInt8Column
!
!*********************************************************

subroutine ReadInt8Column(tableSet, heading, size, start, array)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading
    integer, intent(in) :: size, start
    integer(int8) :: array(:)
    integer(int8), pointer :: i8col(:)
    type(ColumnT) :: pColumn
    integer i, k
!    if (.not. associated(array)) allocate(array(1 : size))  
    k = start
    pColumn = column(tableSet, heading, read)
    i8col => int8Data(pcolumn)
    do i = 1 , size
        k = k  + 1
        array(k) = i8col(i - 1)
    enddo
    call release(pColumn)

end subroutine ReadInt8Column

!**********************************************************
! ReadInt8Column
!
!*********************************************************


subroutine ReadInt16Column(tableSet, heading, size, start, array)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading
    integer, intent(in) :: size, start
    integer(int16) :: array(:)
    integer(int16), pointer :: i16col(:)
    type(ColumnT) :: pColumn
    integer i, k
   
    
    k = start
    pColumn = column(tableSet, heading, read)
    i16col => int16Data(pcolumn)
    do i = 1 , size
        k = k  + 1
        array(k) = i16col(i - 1)
    enddo
    call release(pColumn)

end subroutine ReadInt16Column
!**********************************************************
! ReadInt32Column
!
!*********************************************************
subroutine ReadInt32Column(tableSet, heading, size, start, array)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading
    integer, intent(in) :: size, start
    integer(int32) :: array(:)
    integer(int32), pointer :: i32col(:)
    type(ColumnT) :: pColumn
    integer i, k
      
!   if (.not. associated(array)) allocate(array(1 : size))  
     
    k = start
    pColumn = column(tableSet, heading, read)
    i32col => int32Data(pcolumn)
    do i = 1 , size
        k = k  + 1
        array(k) = i32col(i - 1)
    enddo
    call release(pColumn)

end subroutine ReadInt32Column

!**********************************************************
! ReadReal32Column
!
!*********************************************************
subroutine ReadReal32Column(tableSet, heading, size, start, array)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading
    integer, intent(in) :: size, start
    real(single) :: array(:)
    real(single), pointer :: real32Col(:)
    type(ColumnT) :: pColumn
    integer i, k
!    if (.not. associated(array)) allocate(array(1 : size))  
 
    pColumn = column(tableSet, heading, read)
    real32Col => real32Data(pcolumn)
    k = start
    do i = 1 , size
        k = k + 1   
        array(k) = real32Col(i - 1)
    enddo
    call release(pColumn)

end subroutine ReadReal32Column

!**********************************************************
! ReadReal64Column
!
!*********************************************************
subroutine ReadReal64Column(tableSet, heading, size, start, array)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading
    integer, intent(in) :: size, start
    real(double) :: array(:)
    real(double), pointer :: real64Col(:)
    type(ColumnT) :: pColumn
    integer i, k
!    if (.not. associated(array)) allocate(array(1 : size))  
     
   
    pColumn = column(tableSet, heading, read)
    real64Col => real64Data(pcolumn)
    k = start
    do i = 1 , size
        k = k + 1
        array(k) = real64Col(i - 1)
    end do
    call release(pColumn)

end subroutine ReadReal64Column



!**********************************************************
! ReadInt8Column
!
!*********************************************************


subroutine Read1Int8Column(tableSet, heading, size, array)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading
    integer, intent(in) :: size
    integer(int8) :: array(:)
    integer(int8), pointer :: i8col(:)
    type(ColumnT) :: pColumn
    integer i
     

    pColumn = column(tableSet, heading, read)
    i8col => int8Data(pcolumn)
    do i = 1 , size
       
        array(i) = i8col(i - 1)
    enddo
    call release(pColumn)

end subroutine Read1Int8Column
!**********************************************************
! ReadInt8Column
!
!*********************************************************


subroutine Read1Int16Column(tableSet, heading, size, array)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading
    integer, intent(in) :: size
    integer(int16) :: array(:)
    integer(int16), pointer :: i16col(:)
    type(ColumnT) :: pColumn
    integer i
      
    
    
    pColumn = column(tableSet, heading, read)
    i16col => int16Data(pcolumn)

  
 
    do i = 1 , size
        
        array(i) = i16col(i - 1)
    enddo
    call release(pColumn)

end subroutine Read1Int16Column
!**********************************************************
! ReadInt32Column
!
!*********************************************************
subroutine Read1Int32Column(tableSet, heading, size, array)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading
    integer, intent(in) :: size
    integer(int32) :: array(:)
    integer(int32), pointer :: i32col(:)
    type(ColumnT) :: pColumn
    integer i
     
    
   
    pColumn = column(tableSet, heading, read)
    i32col => int32Data(pcolumn)
   
 

    do i = 1 , size
        
        array(i) = i32col(i - 1)
    enddo
    call release(pColumn)

end subroutine Read1Int32Column
!**********************************************************
! ReadReal32Column
!
!*********************************************************
subroutine Read1Real32Column(tableSet, heading, size, array)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading
    integer, intent(in) :: size
    real(single) :: array(:)
    real(single), pointer :: real32Col(:)
    type(ColumnT) :: pColumn
    integer i
     
 
    pColumn = column(tableSet, heading, read)
    real32Col => real32Data(pcolumn)
  
 
    do i = 1, size
           
        array(i) = real32Col(i - 1)
    enddo
    call release(pColumn)

end subroutine Read1Real32Column
!**********************************************************
! ReadReal64Column
!
!*********************************************************



subroutine Read1Real64Column(tableSet, heading, size, array)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading
    integer, intent(in) :: size
    real(double) :: array(:)
    real(double), pointer :: real64Col(:)
    type(ColumnT) :: pColumn
    integer i
         
    pColumn = column(tableSet, heading, read)
    real64Col => real64Data(pcolumn)
    
    do i = 1 , size
        
        array(i) = real64Col(i - 1)
    end do
    call release(pColumn)

end subroutine Read1Real64Column





!**********************************************************
! ReadCharColumn
!
!*********************************************************
subroutine ReadCharColumn(tableSet, heading, size, start, array)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading
    integer, intent(in) :: size, start
    character(*) :: array(:)
   
    type(ColumnT) :: pColumn
    integer i, k

 
    pColumn = column(tableSet, heading, read)
   
    k = start
    do i = 1 , size
        k = k + 1   
        array(k) = stringcell( pcolumn, i - 1)
    enddo
    call release(pColumn)

end subroutine ReadCharColumn



!**********************************************************
! ReadCharColumn
!
!*********************************************************
subroutine ReadCharColumn1(tableSet, heading, size, array)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading
    integer, intent(in) :: size
    character(*) :: array(:)
   
    type(ColumnT) :: pColumn
    integer i, k

 
    pColumn = column(tableSet, heading, read)
   
    k = 0
    do i = 1 , size
        k = k + 1   
        array(k) = stringcell( pcolumn, i - 1)
    enddo
    call release(pColumn)

end subroutine ReadCharColumn1





!***************************************************
! Add2Int8Column
! For 2-dimensional arrays
!***************************************************
subroutine Add2Int8IndexColumn(tableSet, heading, units, comment, &
        size, array, index, indexOrder, nullValue)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading, units, comment
    integer, intent(in) :: size, index, indexOrder(1 : size)
    integer(int8), intent(in), dimension(:, :) :: array
    integer(int32), intent(in) :: nullValue
    integer(int8), pointer :: int8Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)
 

    if (.not. hasColumn(tableSet, trim(heading))) then
        pColumn = addColumn(tableSet, trim(heading), integer8, trim(units), ndims, trim(comment))
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
   
    

    int8Col => int8Data(pcolumn)
    call setNullValue(pcolumn, nullValue)
    do i = 1 , size
        if (array(indexOrder(i), index) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            int8Col(i - 1) = array(indexOrder(i), index)
        end if
  
    end do
    call release(pcolumn)

end subroutine Add2Int8IndexColumn
!***************************************************
! Add2Int16Column
! For 2-dimensional arrays
!***************************************************
subroutine Add2Int16IndexColumn(tableSet, heading, units, comment, &
        size, array, index, indexOrder, nullValue)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading, units, comment
    integer, intent(in) :: size, index, indexOrder(1 : size)
    integer(int16), intent(in), dimension(:, :) :: array
    integer(int32), intent(in) :: nullValue
    integer(int16), pointer :: int16Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)
 

    if (.not. hasColumn(tableSet, trim(heading))) then
        pColumn = addColumn(tableSet, trim(heading), integer16, trim(units), ndims, trim(comment))
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
   
    

    int16Col => int16Data(pcolumn)
    call setNullValue(pcolumn, nullValue)
    do i = 1 , size
        if (array(indexOrder(i), index) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            int16Col(i - 1) = array(indexOrder(i), index)
        end if
  
    end do
    call release(pcolumn)

end subroutine Add2Int16IndexColumn

!***************************************************
! Add2Real32Column
! For 2-dimensional arrays
!***************************************************
subroutine Add2Real32IndexColumn(tableSet, heading, units, comment, &
        size, array, index, indexOrder, nullValue)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, units, comment
    integer, intent(in) :: size, index, indexOrder(1 : size)
    real(single), intent(in), dimension(:, :) :: array
    real(single), intent(in) :: nullValue
    real(single), pointer :: real32Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)
 



    if (.not. hasColumn(tableSet, trim(heading))) then
        pColumn = addColumn(tableSet, trim(heading), real32, trim(units), ndims, trim(comment))
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
   


    real32Col => real32Data(pcolumn)
  
    do i = 1 , size
        if (array(indexOrder(i), index) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            real32Col(i - 1) = array(indexOrder(i), index)
        end if
  
    enddo
    call release(pcolumn)

end subroutine Add2Real32IndexColumn


!***************************************************
! AddReal64Column
!***************************************************
subroutine AddReal32IndexColumn(tableSet, heading, units, comment, &
        size, array, indexOrder, nullValue, position)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading, units, comment
    integer, intent(in) :: size, indexOrder(:)

    real(single), intent(in), dimension(1 : size) :: array
    real(single), intent(in) :: nullValue
    integer(int32), intent(in), optional :: position

    real(single), pointer :: real32Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)



    if (.not. hasColumn(tableSet, trim(heading))) then
        if (present(position)) then
            pColumn = addColumn(tableSet, trim(heading), real32, trim(units), ndims, trim(comment), position)
        else 
            pColumn = addColumn(tableSet, trim(heading), real32, trim(units), ndims, trim(comment))
        end if
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
   



   
    real32Col => real32Data(pcolumn)

    do i = 1 , size
        if (array(indexOrder(i)) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            real32Col(i - 1) = array(indexOrder(i))
        end if
  
    enddo
    call release(pcolumn)

end subroutine AddReal32IndexColumn




!***************************************************
! AddReal64Column
!***************************************************
subroutine AddReal64IndexColumn(tableSet, heading, units, comment, &
        size, array, indexOrder, nullValue, position)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading, units, comment
    integer, intent(in) :: size, indexOrder(1 : size)

    real(double), intent(in), dimension(:) :: array
    real(single), intent(in) :: nullValue
    integer(int32), intent(in), optional :: position

    real(double), pointer :: real64Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)



    if (.not. hasColumn(tableSet, trim(heading))) then
        if (present(position)) then
            pColumn = addColumn(tableSet, trim(heading), real64, trim(units), ndims, trim(comment), position)
        else 
            pColumn = addColumn(tableSet, trim(heading), real64, trim(units), ndims, trim(comment))
        end if
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
   


    real64Col => real64Data(pcolumn)
!write(*, *) "Addreal64indexcolumn"
    do i = 1 , size
        if (array(indexOrder(i)) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            real64Col(i - 1) = array(indexOrder(i))
        end if
!write(*, *) i, ") index=", indexorder(i), " value=",real64Col(i - 1)
  
    enddo
    call release(pcolumn)

end subroutine AddReal64IndexColumn



!***************************************************
! AddInt8Cololumn
!***************************************************
subroutine AddInt8IndexColumn(tableSet, heading, units, comment, &
        size, indexOrder, array, position)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, units, comment
    integer, intent(in) :: size, indexOrder(1 : size)
    integer(int8), intent(in), dimension(:) :: array
    integer(int32), intent(in), optional :: position


    integer(int8), pointer :: int8Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)

 
   

    if (.not. hasColumn(tableSet, trim(heading))) then
        if (present(position)) then
            pColumn = addColumn(tableSet, trim(heading), integer8, trim(units), ndims, trim(comment), position)
        else 
            pColumn = addColumn(tableSet, trim(heading), integer8, trim(units), ndims, trim(comment))
        end if
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
   



    int8Col => int8Data(pcolumn)

    forall (i = 1 : size) int8Col(i - 1) = array(indexOrder(i))
  
    call release(pcolumn)

end subroutine AddInt8IndexColumn
!***************************************************
! AddInt16Cololumn
!***************************************************
subroutine AddInt16IndexColumn(tableSet, heading, units, comment, &
        size, indexOrder, array, position)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading, units, comment
    integer, intent(in) :: size, indexOrder(1 : size)
    integer(int16), intent(in), dimension(:) :: array
    integer(int32), intent(in), optional :: position

    integer(int16), pointer :: int16Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)

   

    if (.not. hasColumn(tableSet, trim(heading))) then
        if (present(position)) then
            pColumn = addColumn(tableSet, trim(heading), integer16, trim(units), ndims, trim(comment), position)
        else 
            pColumn = addColumn(tableSet, trim(heading), integer16, trim(units), ndims, trim(comment))
        end if
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
   


  
    int16Col => int16Data(pcolumn)

    forall(i = 1: size) int16Col(i - 1) = array(indexOrder(i))

    call release(pcolumn)

end subroutine AddInt16IndexColumn




!***************************************************
! AddInt32Cololumn
!***************************************************
subroutine AddInt32IndexColumn(tableSet, heading, units, comment, &
        size, indexOrder, array, position)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading, units, comment
    integer, intent(in) :: size, indexOrder(1 : size)
    integer(int32), intent(in), dimension(:) :: array
    integer(int32), intent(in), optional :: position

    integer(int32), pointer :: int32Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)

   

    if (.not. hasColumn(tableSet, trim(heading))) then
        if (present(position)) then
            pColumn = addColumn(tableSet, trim(heading), integer32, trim(units), ndims, trim(comment), position)
        else 
            pColumn = addColumn(tableSet, trim(heading), integer32, trim(units), ndims, trim(comment))
        end if
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
   


  
    int32Col => int32Data(pcolumn)

    forall(i = 1: size) int32Col(i - 1) = array(indexOrder(i))

    call release(pcolumn)

end subroutine AddInt32IndexColumn



!***************************************************
! AddCharCololumn
!***************************************************
subroutine AddCharColumn(tableSet, heading, units, comment, &
        size, array, position)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading, units, comment
    integer, intent(in) :: size   
    character(*), intent(in), dimension(:) :: array
    integer(int32), intent(in), optional :: position

    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1)
    character(250) :: string1
    ndims = 0  
    do i = 1, size

        ndims(1) = max(ndims(1), len(trim(array(i))))
    end do
    if (.not. hasColumn(tableSet, trim(heading))) then
        if (present(position)) then
            pColumn = addColumn(tableSet, trim(heading), STRING, trim(units), ndims, trim(comment), position)
        else 
            pColumn = addColumn(tableSet, trim(heading), STRING, trim(units), ndims, trim(comment))
        end if
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
    

    do i = 1, size        
        
        string1 = trim(array(i))
!        string1 = adjustl(string1)
!        string1 = adjustr(string1)
!        write(*, *) i, " string = ", string1
        
        call setStringCell(pcolumn, i - 1, string1(1 : ndims(1)))
    end do

    call release(pcolumn)

end subroutine AddCharColumn



!***************************************************
! AddCharCololumn
!***************************************************
subroutine AddCharIndexColumn(tableSet, heading, units, comment, &
        size, indexOrder, array, position)
    type(TableT), intent(in) :: tableSet
    character(*), intent(in) :: heading, units, comment
    integer, intent(in) :: size, indexOrder(1 : size)
    character(*), intent(in), dimension(:) :: array
    integer(int32), intent(in), optional :: position

    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1)
    character(250) :: string1
    ndims = 0  
    do i = 1, size

        ndims(1) = max(ndims(1), len(trim(array(i))))
    end do
    if (.not. hasColumn(tableSet, trim(heading))) then
        if (present(position)) then
            pColumn = addColumn(tableSet, trim(heading), STRING, trim(units), ndims, trim(comment), position)
        else 
            pColumn = addColumn(tableSet, trim(heading), STRING, trim(units), ndims, trim(comment))
        end if
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
    

    do i = 1, size        
        
        string1 = trim(array(indexOrder(i)))
!        string1 = adjustl(string1)
!        string1 = adjustr(string1)
!        write(*, *) i, " string = ", string1
        
        call setStringCell(pcolumn, i - 1, string1(1 : ndims(1)))
    end do

    call release(pcolumn)

end subroutine AddCharIndexColumn

!******************************************
! Subroutine SaveInt8Image
! Creates a file to store the image
!******************************************
subroutine SaveInt8Image(image, nx, ny, fileName, heading)
    integer(int16), intent(in) :: nx, ny    
 
    integer(int8), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: image
    character(*), intent(in) :: fileName
    character(*), intent(in) :: heading
    type(DataSetT)       :: imageSet 
    integer(int8), dimension(:, :), pointer :: imageP
    type(ArrayT)         :: imageArray         ! array for background
    
    
    imageSet = DataSet(fileName, Create)
    if (hasBlock(imageSet, "PRIMARY")) then
        call deleteBlock( imageSet, "PRIMARY" );
    end if

    if (heading .eq. "test") then

    end if  

    imageArray = addArray(imageSet, "PRIMARY", Integer8, &
                 shape(image))

    imageP => int8Array2Data(imageArray)

    imageP = image

    call release(imageArray)

    call release(imageSet)


end subroutine SaveInt8Image

!******************************************
! Subroutine SaveInt16Image
! Creates a file to store the image
!******************************************
subroutine SaveInt16Image(levelImage, nx, ny, fileName, heading)
    integer(int16), intent(in) :: nx, ny    
 
    integer(int16), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: levelImage
    character(*), intent(in) :: fileName
    character(*), intent(in) :: heading
    type(DataSetT)       :: levelImageSet 
    integer(int16), dimension(:, :), pointer :: levelImageP
    type(ArrayT)         :: levelImageArray         ! array for background
    
    
    levelImageSet = DataSet(fileName, Create)
    if (hasBlock(levelImageSet, "PRIMARY")) then
        call deleteBlock( levelimageSet, "PRIMARY" );
    end if

  

    if (heading .eq. "test") then

    end if  


    levelImageArray = addArray(levelImageSet, "PRIMARY", Integer16, &
                 shape(levelImage))

    levelImageP => int16Array2Data(levelImageArray)

    levelImageP = levelImage

    call release(levelImageArray)

    call release(levelImageSet)


end subroutine SaveInt16Image
!****************************************************
! Subroutine RealImage
!****************************************************
subroutine SaveRealImage1(image, nx, ny, fileName, heading)

    integer (int16), intent(in) :: nx, ny    
 
    real (single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: image
    character(*), intent(in) :: fileName
    character(*), intent(in) :: heading
    type(DataSetT)       :: imageSet 
    real (single), dimension(:, :), pointer :: imageP
    type(ArrayT)         :: imageArray         ! array for background
   
 !write(*, *) "omutility_mod saverealimage1 line 1300, saving to file ", trim(fileName)
    imageSet = DataSet(trim(fileName), Create)

    if (hasBlock(imageSet, "PRIMARY")) then
        call deleteBlock( imageSet, "PRIMARY" );
    end if



    if (heading .eq. "test") then

    end if  




    imageArray = addArray(imageSet, "PRIMARY", real32, shape(image))

    imageP => real32Array2Data(imageArray)

    imageP = image

    call release(imageArray)

    call release(imageSet)
 


!write(*, *) "Leaving saverealimage1"
   
end subroutine SaveRealImage1

!****************************************************
!
! Subroutine RealImage
!
!****************************************************
subroutine SaveRealImage2(image, fileName, heading)

     
 
    real (single) :: image(:, :)
    character(*), intent(in) :: fileName
    character(*), intent(in) :: heading
    type(DataSetT)       :: imageSet 
    real (single), dimension(:, :), pointer :: imageP
    type(ArrayT)         :: imageArray         ! array for background
   
 
    imageSet = DataSet(fileName, Create)
    if (hasBlock(imageSet, "PRIMARY")) then
        call deleteBlock( imageSet, "PRIMARY" );
    end if



    if (heading .eq. "test") then

    end if  




    imageArray = addArray(imageSet, "PRIMARY", real32, shape(image))

    imageP => real32Array2Data(imageArray)

    imageP = image

    call release(imageArray)

    call release(imageSet)

end subroutine SaveRealImage2





!*******************************************
! subroutine GetImage
! Gets the 2-d image from the fits file
!*******************************************
subroutine GetImage(dataSet, nx, ny, image) 
!    type(dataSetT), intent(in) :: dataSet 
! VNY test 20.06.2003
    type(dataSetT) :: dataSet
! 
    real(single), pointer, dimension(: , :) :: image
    integer(int16), intent(out) :: nx, ny    
    integer(int8), dimension(:, :),  pointer :: int8ImageP     ! Integer8 array pointer.
    integer(int32), dimension(:, :), pointer :: int32ImageP    ! Integer32 array pointer.
    integer(int16), dimension(:, :), pointer :: int16ImageP    ! Integer16 array pointer.
    real(single), dimension(:, :),   pointer :: real32ImageP   ! Real32 array pointer. 
    real(double), dimension(:, :),   pointer :: real64ImageP   ! Real64 array pointer.
    integer :: dim(2)
    type(ArrayT) :: inArray
    integer :: nDim
    
    inArray = array(dataSet, "PRIMARY", modify)
!
    nDim  = numberOfDimensions(inArray)             ! How many dimensions?
    ! If it's not an image then crash out
    if (nDim .ne. 2) call fatal('notImage','input file not an image')
    ! Get the number of bins.
    dim = dimensions(inArray)
    nx = dim(1)
    ny = dim(2)
    if (nx <= 0) call fatal("Bad image", "nx <=0")
    if (ny <= 0) call fatal("Bad image", "ny <=0")


    if (associated(image)) deallocate(image)
    image => Allocate2DSingle(0,  nx - 1,  0, ny - 1, "image")
    ! Convert the image to real, if necessary
    select case(arrayDataType(inArray))
        case(integer8)   
            int8ImageP => int8Array2Data(inArray)
            image = real(int8ImageP)
        case(integer16)
            int16ImageP => int16Array2Data(inArray)
            image = real(int16ImageP)    
        case(integer32) 
            int32ImageP => int32Array2Data(inArray)
            image = real(int32ImageP)
        case(real32)
            real32ImageP => real32Array2Data(inArray)
            image = real32ImageP       
        case(real64)
            real64ImageP => real64Array2Data(inArray)
            image = real(real64ImageP)
    end select
    call release(inArray)
end subroutine GetImage

!**********************************
! Function GetQualityArray
!          Purpose: to return the quality array as boolean values
!**********************************
function GetQualityArray(dataSet, nx, ny) result(qualityImage)
    type(dataSetT), intent(in) :: dataSet
    integer(int16), intent(in) :: nx, ny
    logical(boolean), pointer :: qualityImage(:, :)
    type(ArrayT) :: inArray
   
    integer(int8), dimension(:, :),   pointer :: int8ImageP  => null()
    integer(int16), dimension(:, :),  pointer :: int16ImageP => null()
    integer :: i, j
  

    allocate(qualityImage(0 : nx - 1, 0 : ny - 1))

    ! Get a handle and pointer on the quality array and extract the quality data.
    inArray = array(dataSet, "QUALITY", read)
 

    select case(arrayDataType(inArray))
        case(integer8)   
            int8ImageP => int8Array2Data(inArray)
            
            where (int8ImageP(0 : nx - 1, 0 : ny - 1) .ne. 0) 
                qualityImage = .true.
            elsewhere
                qualityImage = .false.      
            end where
  

        case(integer16)
           
            int16ImageP => int16Array2Data(inArray)
            do j = 0, ny - 1
                do i = 0, nx - 1
                    if ( int16ImageP(i, j) .eq. 32767) then
                        qualityImage(i, j) = .true.
                   
                    elseif(int16ImageP(i, j) .eq. 0) then
                        qualityImage(i, j) = .false.
                   

                    else
                        if (btest(int16ImageP(i, j), 0) ) then
                            qualityImage(i, j) = .true.
                        else 
                            qualityImage(i, j) = .false.
                        end if
                
                                      
                    end if
                end do
            end do
      end select



    call release(inArray)

end function GetQualityArray


!**************************************************************
! Function GetQualityArrayInt16
!          Purpose: to return the original quality array values
!**************************************************************
function GetQualityArrayInt16(dataSet, nx, ny) result(int16ImageP)
    type(dataSetT), intent(in) :: dataSet
    integer(int16), intent(in) :: nx, ny
    integer(int16), pointer :: int16ImageP(:, :)
    type(ArrayT) :: inArray
   
    integer(int8), dimension(:, :),   pointer :: int8ImageP  => null()
    integer :: i, j
  
    ! Get a handle and pointer on the quality array and extract the quality data.
    inArray = array(dataSet, "QUALITY", read)

    select case(arrayDataType(inArray))
        case(integer8)   
            int8ImageP => int8Array2Data(inArray)
            
            allocate(int16ImageP(0 : nx - 1, 0 : ny - 1))

            do j = 0, ny - 1
                do i = 0, nx - 1
                   int16ImageP(i, j) = int8ImageP(i, j)
                end do
            end do

        case(integer16)
            int16ImageP => int16Array2Data(inArray)

    end select



    call release(inArray)

end function GetQualityArrayInt16

!****************************************************************************
! Returns an array containing the data in the named column of the given table
!
!****************************************************************************
subroutine  LoadReal32Column(inTableH, name, data)
    type(TableT), intent(in) :: inTableH
    character(*), intent(in) :: name
    real(single), intent(out) :: data(:)

    type(ColumnT)         :: columnH
    real(single), pointer :: dataPtr(:) ! Magnitude error pointer.
    integer(int32)        :: i, sz
     

     
    if (hasColumn(inTableH, name)) then
        columnH = column(inTableH, name)
        dataPtr => real32Data(columnH)
        sz = numberOfRows(inTableH)

        do i = 1, sz
            data(i) = dataPtr(i - 1)
           
        end do
        call release(columnH)
    else
        call warning("No such data column ", name)


    end if
end subroutine LoadReal32Column

!******************************************************************************
! Returns an array containing the data in the named column of the given table
!******************************************************************************
subroutine LoadReal64Column(inTableH, name, data)
    type(TableT), intent(in) :: inTableH
    character(*), intent(in) :: name
    real(double), intent(out) :: data(:)

    type(ColumnT)         :: columnH
    real(double), pointer :: dataPtr(:) ! Magnitude error pointer.
    integer(int32)        :: i, sz
           
    if (hasColumn(inTableH, name)) then
        columnH = column(inTableH, name)
        dataPtr => real64Data(columnH)
        sz = numberOfRows(inTableH)

        do i = 1, sz
            data(i) = dataPtr(i - 1)
        end do
        call release(columnH)
    end if
end subroutine LoadReal64Column


!***************************************************************************
!
! Returns an array containing the data in the named column of the given table
!***************************************************************************
subroutine LoadInt8Column(inTableH, name, data) 
    type(TableT), intent(in) :: inTableH
    character(*), intent(in) :: name
    integer(int8), intent(out) :: data(:)

    type(ColumnT)         :: columnH
    integer(int8), pointer :: dataPtr(:) ! Magnitude error pointer.
    integer(int32)        :: i, sz
     
    if (hasColumn(inTableH, name)) then
        columnH = column(inTableH, name)
        dataPtr => int8Data(columnH)
        sz = numberOfRows(inTableH)


        do i = 1, sz
            data(i) = dataPtr(i - 1)

        end do
        call release(columnH)

    else

        call warning("No such data column ", name)

    end if
end subroutine LoadInt8Column




!***************************************************************************
!
! Returns an array containing the data in the named column of the given table
!***************************************************************************
subroutine LoadInt16Column(inTableH, name, data) 
    type(TableT), intent(in) :: inTableH
    character(*), intent(in) :: name
    integer(int16), intent(out) :: data(:)

    type(ColumnT)         :: columnH
    integer(int16), pointer :: dataPtr(:) ! Magnitude error pointer.
    integer(int32)        :: i, sz
     
    if (hasColumn(inTableH, name)) then
        columnH = column(inTableH, name)
        dataPtr => int16Data(columnH)
        sz = numberOfRows(inTableH)

        do i = 1, sz
            data(i) = dataPtr(i - 1)
        end do
        call release(columnH)

    else

        call warning("No such data column ", name)


    end if
end subroutine LoadInt16Column
!***************************************************************************
!
! Returns an array containing the data in the named column of the given table
!***************************************************************************
subroutine LoadInt32Column(inTableH, name, data) 
    type(TableT), intent(in) :: inTableH
    character(*), intent(in) :: name
    integer(int32), intent(out) :: data(:)

    type(ColumnT)         :: columnH
    integer(int32), pointer :: dataPtr(:) ! Magnitude error pointer.
    integer(int32)        :: i, sz
     
    if (hasColumn(inTableH, name)) then
        columnH = column(inTableH, name)
        dataPtr => int32Data(columnH)
        sz = numberOfRows(inTableH)

        do i = 1, sz
            data(i) = dataPtr(i - 1)
        end do
        call release(columnH)
    
    else

        call warning("No such data column ", name)

    end if
end subroutine LoadInt32Column

!**********************************************************************
!
!
!**********************************************************************
subroutine LoadRealVectorColumn(tableSet, columnName, data, nrows, nCols)
 
    type(tableT), intent(in) :: tableSet
    character(*) :: columnName
    real(single), pointer :: data(:, :)
    integer, intent(out) :: nrows, nCols
   
   
    integer :: i, j
  
 
    type(ColumnT) :: col
  
    real(single), dimension(:, :), pointer :: c


    if (hasColumn(tableSet, trim(columnName))) then
       
        col = column(tableSet, trim(columnName), read)
    end if

    c => real32VectorData(col)
    
    

    nrows = size(c, 2)
    ncols = size(c, 1)  

    if (associated(data)) deallocate(data)
    data => allocate2dSingle(1, nrows, 1, ncols, " data")

    do i = 1, nrows
  
     
        do j = 1, nCols
       

   
            data(i, j) = c(j - 1, i - 1) 
            
        end do
    end do

    call release(col)


 
   
end subroutine  LoadRealVectorColumn


!**********************************************************************
!
!
!**********************************************************************
subroutine LoadDoubleVectorColumn(tableSet, columnName, data, nrows, nCols)
 
    type(tableT), intent(in) :: tableSet
    character(*) :: columnName
    real(double), pointer :: data(:, :)
    integer, intent(out) :: nrows, nCols
   
   
    integer :: i, j
  
 
    type(ColumnT) :: col
  
    real(double), dimension(:, :), pointer :: c


    if (hasColumn(tableSet, trim(columnName))) then
       
        col = column(tableSet, trim(columnName), read)
    end if

    c => real64VectorData(col)
    
    

    nrows = size(c, 1)
    ncols = size(c, 2)  

    if (associated(data)) deallocate(data)
    data => allocate2dDouble(1, nrows, 1, ncols, " data")

    do i = 1, nrows
  
     
        do j = 1, nCols
       

   
            data(i, j) = c(j - 1, i - 1) 
            
        end do
    end do

    call release(col)


 
   
end subroutine  LoadDoubleVectorColumn



!**********************************************************************
!
!
!**********************************************************************


subroutine LoadInt8VectorColumn(tableSet, columnName, data, nrows, nCols)
 
    type(tableT), intent(in) :: tableSet
    character(*) :: columnName
    integer(int8), pointer :: data(:, :)
    integer, intent(out) :: nrows, nCols
   
   
    integer :: i, j
  
 
    type(ColumnT) :: col
  
    integer(int8), dimension(:, :), pointer :: c

 


    if (hasColumn(tableSet, trim(columnName))) then
       
        col = column(tableSet, trim(columnName), read)
    end if

    c => int8VectorData(col)
    
    

    nrows = size(c, 1)
    ncols = size(c, 2)  

    if (associated(data)) deallocate(data)
    data => allocate2dint8(1, nrows, 1, ncols, " data")

    do i = 1, nrows
  
     
        do j = 1, nCols
       

   
            data(i, j) = c(j - 1, i - 1) 
            
        end do
    end do

    call release(col)
   

 
   
end subroutine  LoadInt8VectorColumn

!**********************************************************************
!
!
!**********************************************************************


subroutine LoadInt16VectorColumn(tableSet, columnName, data, nrows, nCols)
 
    type(tableT), intent(in) :: tableSet
    character(*) :: columnName
    integer(int16), pointer :: data(:, :)
    integer, intent(out) :: nrows, nCols
   
   
    integer :: i, j
  
 
    type(ColumnT) :: col
  
    integer(int16), dimension(:, :), pointer :: c

 


    if (hasColumn(tableSet, trim(columnName))) then
       
        col = column(tableSet, trim(columnName), read)
    end if

    c => int16VectorData(col)
    
    

    nrows = size(c, 1)
    ncols = size(c, 2)  

    if (associated(data)) deallocate(data)
    data => allocate2dint16(1, nrows, 1, ncols, " data")

    do i = 1, nrows
  
     
        do j = 1, nCols
       

   
            data(i, j) = c(j - 1, i - 1) 
            
        end do
    end do

    call release(col)
   

 
   
end subroutine  LoadInt16VectorColumn
!**********************************************************************
!
!
!**********************************************************************


subroutine LoadInt32VectorColumn(tableSet, columnName, data, nrows, nCols)
 
    type(tableT), intent(in) :: tableSet
    character(*) :: columnName
    integer(int32), pointer :: data(:, :)
    integer, intent(out) :: nrows, nCols
   
   
    integer :: i, j 
 
    type(ColumnT) :: col
  
    integer(int32), dimension(:, :), pointer :: c

 


    if (hasColumn(tableSet, trim(columnName))) then
       
        col = column(tableSet, trim(columnName), read)
    end if

    c => int32VectorData(col)
    
    

    nrows = size(c, 1)
    ncols = size(c, 2)  

    if (associated(data)) deallocate(data)
    data => allocate2dint32(1, nrows, 1, ncols, " data")

    do i = 1, nrows
  
     
        do j = 1, nCols
       

   
            data(i, j) = c(j - 1, i - 1) 
            
        end do
    end do

    call release(col)
   

 
   
end subroutine  LoadInt32VectorColumn




!**************************************************************************
!
!
!**************************************************************************

subroutine addRealVectorColumn1(tab, data, nData, nCols, colIndex, storeIndex, &
      heading, units, comment, nullValue, position)
 
    type(tableT), intent(inout) :: tab
    integer, intent(in) :: nData, nCols
    real(single), pointer :: data(:, :)
    integer, intent(in), optional :: colIndex(:)
    integer, intent(in), optional :: storeIndex(:)
    character(*), intent(in) :: heading, units, comment
    real(single), intent(in) :: nullValue
    integer, intent(in), optional :: position 
    integer :: i, j, k, l
 
  
 
    type(ColumnT) :: col
    real(single), dimension(:), pointer :: c1
    integer, dimension(1) :: s 
    real(single), pointer :: tempData(:, :) => null()



    if (ncols .eq. 0 .or. ndata .eq. 0) return
    s(1) = nCols


    tempData => allocate2dSingle(1, nData, 1, nCols, "tempData")


    if (present(storeindex)) then
        do i = 1, nData
            k = storeIndex(i)

            do j = 1, nCols
                if (present(colindex)) then
                    l = colIndex(j)
                else
                    l = j
                end if
                tempData(i, j) = data(k, l)
            end do                       
        end do
    else

        
        do i = 1, nData
         

            do j = 1, nCols
               if (present(colindex)) then
                    l = colIndex(j)
                else
                    l = j
                end if
          
              
                  
                tempData(i, j) = data(i, l)
            end do                       
        end do
    end if   




    if (present(position)) then
        Col = addColumn(tab, trim(heading), REAL32, trim(units), s, trim(comment), position)
    else 
        col = addColumn(tab, trim(heading), REAL32, trim(units), s, trim(comment))
    end if
 


    do i = 1, nData
      
        c1 => Real32VectorData(col, i - 1)
 
        do j = 1, nCols
       
            if (tempData(i, j) .eq. nullValue) then
                call setToNull(col, i - 1, j - 1)

            else

                c1(j - 1) = tempData(i, j)
            end if         


        end do

    end do


    call release(col)
    deallocate(tempdata)
    nullify(tempdata)  

end subroutine addRealVectorColumn1

!**************************************************************************
!
!
!**************************************************************************

subroutine addDoubleVectorColumn1(tab, data, nData, nCols, colIndex, storeIndex, &
      heading, units, comment, nullValue, position)
 
    type(tableT), intent(inout) :: tab
    integer, intent(in) :: nData, nCols
    real(double), pointer :: data(:, :)
    integer, intent(in), optional :: colIndex(:)
    integer, intent(in), optional :: storeIndex(:)
    character(*), intent(in) :: heading, units, comment
    real(single), intent(in) :: nullValue
    integer, intent(in), optional :: position 
    integer :: i, j, k, l
 
  
 
    type(ColumnT) :: col
    real(double), dimension(:), pointer :: c1
    integer, dimension(1) :: s 
    real(double), pointer :: tempData(:, :) => null()

    if (ncols .eq. 0 .or. ndata .eq. 0) return


    s(1) = nCols


    tempData => allocate2dDouble(1, nData, 1, nCols, "tempData")

    if (present(storeindex)) then
        do i = 1, nData
            k = storeIndex(i)

            do j = 1, nCols
               
                if (present(colindex)) then
                    l = colIndex(j)
                else
                    l = j
                end if
            
                tempData(i, j) = data(k, l)
            end do                       
        end do
    else

        do i = 1, nData
         

            do j = 1, nCols
              
                if (present(colindex)) then
                    l = colIndex(j)
                else
                    l = j
                end if
    



                tempData(i, j) = data(i, l)
            end do                       
        end do
    end if   

    if (present(position)) then
        Col = addColumn(tab, trim(heading), REAL64, trim(units), s, trim(comment), position)
    else 
        col = addColumn(tab, trim(heading), REAL64, trim(units), s, trim(comment))
    end if
 


  

    do i = 1, nData
      
        c1 => Real64VectorData(col, i - 1)
  

        do j = 1, nCols
       
            if (tempData(i, j) .eq. nullValue) then
                call setToNull(col, i - 1, j - 1)

            else

                c1(j - 1) = tempData(i, j)
            end if         

        end do
    end do
    call release(col)
    deallocate(tempdata)
    nullify(tempdata)  

end subroutine addDoubleVectorColumn1



!**************************************************************************
!
!
!**************************************************************************

   


subroutine AddInt8VectorColumn1(tab, data, nData, nCols, colIndex, storeIndex, &
      heading, units, comment, nullValue, position)
 
     type(tableT), intent(inout) :: tab
 
 
    integer, intent(in) :: nData, nCols
 
    integer(int8), pointer :: data(:, :)
    integer, intent(in), optional :: colIndex(:)
    integer, intent(in), optional :: storeIndex(:)
    character(*), intent(in) :: heading, units, comment
    integer(int8), intent(in) :: nullValue
    integer, intent(in), optional :: position 
    integer :: i, j, k, l
 
  
 
    type(ColumnT) :: col
    integer(int8), dimension(:), pointer :: c1
    integer, dimension(1) :: s 
    integer(int8), pointer :: tempData(:, :) => null()

    if (ncols .eq. 0 .or. ndata .eq. 0) return

    s(1) = nCols


    tempData => allocate2dInt8(1, nData, 1, nCols, "tempData")

    if (present(storeindex)) then
        do i = 1, nData
            k = storeIndex(i)

            do j = 1, nCols
                if (present(colindex)) then
                    l = colIndex(j)
                else
                    l = j
                end if
    
                tempData(i, j) = data(k, l)
            end do                       
        end do
    else

        do i = 1, nData
         

            do j = 1, nCols
                if (present(colindex)) then
                    l = colIndex(j)
                else
                    l = j
                end if
    
                tempData(i, j) = data(i, l)
            end do                       
        end do
    end if   

    if (present(position)) then
        Col = addColumn(tab, trim(heading), INTEGER8, trim(units), s, trim(comment), position)
    else 
        col = addColumn(tab, trim(heading), INTEGER8, trim(units), s, trim(comment))
    end if
 


!    call setNullValue(col, int(nullValue)) 

    call setNullValue(col, int(huge(0_int8), int32)) 


    do i = 1, nData
      
        c1 => Int8VectorData(col, i - 1)
  

        do j = 1, nCols
       
            if (tempData(i, j) .eq. nullValue) then
                
                call setToNull(col, i - 1, j - 1)

            else

                c1(j - 1) = tempData(i, j)
            end if         

        end do
    end do
    call release(col)
    deallocate(tempdata)
    nullify(tempdata)  

end subroutine AddInt8VectorColumn1


!**************************************************************************
!
!
!**************************************************************************

   


subroutine AddInt16VectorColumn1(tab, data, nData, nCols, colIndex, storeIndex, &
      heading, units, comment, nullValue, position)
 
    type(tableT), intent(inout) :: tab
   integer, intent(in) :: nData, nCols
 
    integer(int8), pointer :: data(:, :)
  
    integer, intent(in), optional :: colIndex(:)
    integer, intent(in), optional :: storeIndex(:)
    character(*), intent(in) :: heading, units, comment
    integer(int16), intent(in) :: nullValue
    integer, intent(in), optional :: position 
    integer :: i, j, k, l
 
  
 
    type(ColumnT) :: col
    integer(int16), dimension(:), pointer :: c1
    integer, dimension(1) :: s 
    integer(int16), pointer :: tempData(:, :) => null()

    if (ncols .eq. 0 .or. ndata .eq. 0) return

    s(1) = nCols


    tempData => allocate2dInt16(1, nData, 1, nCols, "tempData")

    if (present(storeindex)) then
        do i = 1, nData
            k = storeIndex(i)

            do j = 1, nCols
                if (present(colindex)) then
                    l = colIndex(j)
                else
                    l = j
                end if
    
                tempData(i, j) = data(k, l)
            end do                       
        end do
    else

        do i = 1, nData
         

            do j = 1, nCols
                    if (present(colindex)) then
                    l = colIndex(j)
                else
                    l = j
                end if
    
                tempData(i, j) = data(i, l)
            end do                       
        end do
    end if   

    if (present(position)) then
        Col = addColumn(tab, trim(heading), INTEGER16, trim(units), s, trim(comment), position)
    else 
        col = addColumn(tab, trim(heading), INTEGER16, trim(units), s, trim(comment))
    end if
 


!  call setNullValue(col, int(nullValue)) 


 call setNullValue(col, int(huge(0_int16), int32)) 

    do i = 1, nData
      
        c1 => Int16VectorData(col, i - 1)
  

        do j = 1, nCols
       
            if (tempData(i, j) .eq. nullValue) then
                call setToNull(col, i - 1, j - 1)

            else

                c1(j - 1) = tempData(i, j)
            end if         

        end do
    end do
    call release(col)
    deallocate(tempdata)
    nullify(tempdata)  

end subroutine AddInt16VectorColumn1
!**************************************************************************
!
!
!**************************************************************************

   


subroutine AddInt32VectorColumn1(tab, data, nData, nCols, colIndex, storeIndex, &
      heading, units, comment, nullValue, position)
 
    type(tableT), intent(inout) :: tab
   integer, intent(in) :: nData, nCols
 
    integer(int32), pointer :: data(:, :)
    
    integer, intent(in), optional :: colIndex(:)
    integer, intent(in), optional :: storeIndex(:)
    character(*), intent(in) :: heading, units, comment
    integer(int32), intent(in) :: nullValue
    integer, intent(in), optional :: position 
    integer :: i, j, k, l
 
  
 
    type(ColumnT) :: col
    integer(int32), dimension(:), pointer :: c1
    integer, dimension(1) :: s 
    integer(int32), pointer :: tempData(:, :) => null()

    
    if (ncols .eq. 0 .or. ndata .eq. 0) return

    s(1) = nCols


    tempData => allocate2dInt32(1, nData, 1, nCols, "tempData")

    if (present(storeindex)) then
        do i = 1, nData
            k = storeIndex(i)

            do j = 1, nCols
                    if (present(colindex)) then
                    l = colIndex(j)
                else
                    l = j
                end if
    
                tempData(i, j) = data(k, l)
            end do                       
        end do
    else

        do i = 1, nData
         

            do j = 1, nCols
                    if (present(colindex)) then
                    l = colIndex(j)
                else
                    l = j
                end if
    
                tempData(i, j) = data(i, l)
            end do                       
        end do
    end if   

    if (present(position)) then
        Col = addColumn(tab, trim(heading), INTEGER32, trim(units), s, trim(comment), position)
    else 
        col = addColumn(tab, trim(heading), INTEGER32, trim(units), s, trim(comment))
    end if
 


!    call setNullValue(col, int(nullValue)) 
  call setNullValue(col, huge(0_int32)) 


    do i = 1, nData
      
        c1 => Int32VectorData(col, i - 1)
  

        do j = 1, nCols
       
            if (tempData(i, j) .eq. nullValue) then
                call setToNull(col, i - 1, j - 1)

            else

                c1(j - 1) = tempData(i, j)
            end if         

        end do
    end do
    call release(col)
    deallocate(tempdata)
    nullify(tempdata)  

end subroutine AddInt32VectorColumn1





!***************************************
! subroutine ConvertRA
!***************************************
subroutine ConvertRA(ra, hours, minutes, seconds)
    real(double), intent(in) :: ra
    integer, intent(out) ::  hours, minutes
    real, intent(out) :: seconds
    double precision ra1, d

    ra1 = ra / 15.0d0  ! RA in hours
 
    if (ra1 .lt.  0.0d0) ra1 = ra1 + 24.0d0
    if (ra1 .gt. 24.0d0) ra1 = ra1 - 24.0d0
    hours = int(ra1)

    d = (ra1 - dble(hours)) * 60.0d0
    minutes = int(d + 0.00000001d0)
    seconds = (d - dble(minutes)) * 60.0d0
!    write(*, *) "Converted Ra = ", hours, minutes, seconds
  
end subroutine ConvertRA

!***************************************
! subroutine ConvertDec
!***************************************
subroutine ConvertDec(dec, degrees, minutes, seconds)
    real(double), intent(in) :: dec
    integer, intent(out) :: degrees, minutes
    real, intent(out) :: seconds
    double precision dec1
    double precision ::  d  
    degrees = int(dec)    
    dec1 = abs(dec)
    d = (dec1 - dble(abs(degrees))) * 60.0d0
    minutes = int(d + 0.00000001d0)
    seconds = (d - dble(minutes)) * 60.0d0

    if (dec .lt. 0.0d0 .and. degrees .eq. 0) minutes = -minutes
    
    if (dec .lt. 0.0d0 .and. degrees .eq. 0 .and. minutes .eq. 0) seconds = -seconds
 
!    write(*, *) "converted dec = ", degrees, minutes, seconds 

end subroutine ConvertDec

!***************************************************************************************
!
!
!**************************************************************************************
subroutine CreateRADecStrings(ra, dec, rastring, decstring)

    real(double), intent(in) :: ra, dec
    character(*), intent(out) :: rastring, decstring
    integer :: r1, r2, d1, d2
    real(single) :: r3, d3
    character(10) :: s1, s2, s3
   
    call ConvertRA(ra, r1, r2, r3)

    !******************************************
    ! Create the RA string
    !******************************************
    if (r1 < 10) then    ! Hours
        write(s1, '(a1, i1)')  "0", r1
    else
        write(s1, '(i2)') r1
    end if 

    if (r2 < 10) then    ! Minutes
        write(s2, '(a1, i1)')  "0", r2
    else
        write(s2, '(i2)') r2
    end if 

    if (r3 < 9.95) then    ! Seconds
        write(s3, '(a1, f3.1)')  "0", r3
    else
        write(s3, '(f4.1)') r3
    end if 
                       
!    raString = trim(s1) // ":" // trim(s2) // ":" // trim(s3) 

  raString = trim(s1) // "h " // trim(s2) // "m " // trim(s3) // "s" 

    !******************************************
    ! Create the dec string
    !******************************************
    call ConvertDec(dec, d1, d2, d3)
    if (d1 < 0 .or. d2 < 0 .or. d3 < 0.0) then
        if (d1 > -10) then   ! Degrees
            write(s1, '(a2, i1)') "-0", abs(d1)
        else 
                write(s1, '(a1, i2)') "-", abs(d1)
        end if
    else
        if (d1 < 10) then
            write(s1, '(a2, i1)') "+0", d1
        else 
            write(s1, '(a1, i2)') "+", d1
        end if
    end if

    d2 = abs(d2)   ! Minutes
    if (d2 < 10) then
        write(s2, '(a1, i1)') "0", d2
    else 
        write(s2, '(i2)') d2
    end if
          
    d3 = abs(d3)   ! Seconds          
    if (d3 < 9.95) then
        write(s3, '(a1, f3.1)') "0", d3  !
    else 
        write(s3, '(f4.1)') d3
    end if
           
    !*********************************************
    ! Plot the source number and its coordinates
    !********************************************* 
!    decstring = trim(s1) // ":" // trim(s2) // ":" // trim(s3)          
     decstring = trim(s1) // "d " // trim(s2) // "m " // trim(s3) // "s"          
          



end subroutine CreateRADecStrings
!***************************************************************************************
!
!
!**************************************************************************************
subroutine addradecstringcolumns(srcTable, n, ra, dec, rasortindex, position)
    type(TableT), intent(inout) :: srcTable
    integer, intent(in) :: n
    real(double) :: ra(1 : n), dec(1 : n)
    integer, optional :: rasortIndex(1 : n)
    integer, intent(in), optional :: position
    character(20), dimension(:), pointer :: raString => null(), decstring => null()
    integer :: i
    allocate(rastring(1 : n))
    allocate(decstring(1 : n))

   
!write(*, *) "addradecstringcolumns"
    do i = 1, n

        call CreateRADecStrings(ra(i), dec(i), rastring(i), decstring(i))
!        write(*, *) i, ") ra=", ra(i), " ", trim(rastring(i)), " dec=", dec(i), " ", trim(decstring(i))
    end do
! do i = 1, n

!        k = rasortindex(i)
!        write(*, *) i, " ", ra(k), " ", dec(k), " ", trim(rastring(k)), " ", trim(decstring(k))
!    end do

 


 
    if(present(rasortindex)) then

        if (present(position)) then



            call AddCharIndexColumn(srcTable, "DEC_DMS", " ", "Deg-Min-Sec Declination (J2000)", n, &
                RASortIndex, decstring, position)


            call AddCharIndexColumn(srcTable, "RA_HMS", " ", "Hr-Min-Sec Right Ascension (J2000)", n, &
                RASortIndex, rastring, position)
 
       

        else
            call AddCharIndexColumn(srcTable, "RA_HMS", " ", "Hr-Min-Sec Right Ascension (J2000)" , n, &
                RASortIndex, rastring)
 
            call AddCharIndexColumn(srcTable, "DEC_DMS", " ", "Deg-Min-Sec Declination (J2000)" , n, &
                RASortIndex, decstring)
        end if  
    else
       
 
   

        if (present(position)) then


 
            call AddCharColumn(srcTable, "DEC_DMS", " ", "Deg-Min-Sec Declination (J2000)" , n, &
                decstring, position)
 
   

            call AddCharColumn(srcTable, "RA_HMS", " ",  "Hr-Min-Sec Right Ascension (J2000)", n, &
                rastring, position)
 
        
 
       else
            call AddCharColumn(srcTable, "RA_HMS", " ",  "Hr-Min-Sec Right Ascension (J2000)", n, &
                rastring)
 
            call AddCharColumn(srcTable, "DEC_DMS", " ", "Deg-Min-Sec Declination (J2000)" , n, &
                decstring)
        end if  

    end if


    deallocate(rastring, decstring)

end subroutine addradecstringcolumns




 
!*********************************************
! CreateRegionFile
!*********************************************
subroutine CreateRegionFile11(nSources, source, regionFile, plotNumbers)
    integer(int32), intent(in) :: nSources

    type(sourceParameters), pointer :: source(:)

    character(*), intent(in) :: regionFile
    logical, optional :: plotNumbers
    logical :: plotNos  
    integer(int32) :: i
    integer(int32), parameter :: ioUnit = 2
    character(10) :: colour
    character(10) :: number  
    real(single) :: minWidth, theta
 

    if ( present(plotNumbers)) then
        plotNos = plotNumbers
    else
        plotNos = .true.
    end if


    open(unit = ioUnit, file = regionfile, status = 'replace')
!    write(ioUnit,"(a)") '# Region file format: DS9 version 3.0'
!    write(ioUnit,"(a)") 'global color=green font="helvetica &
!    &10 normal" edit=0 move=0 delete=1 include=1 fixed=0'

   
    write(iounit, *) "global select=0 edit=0 move=0 delete=1"


    minWidth = 1.0 / 3600.0

    do i = 1, nSources

        write(number, '(i5)') i
        number = adjustl(number)




        colour = RegionColour(source(i)%confusionFlag, source(i)%qualityFlag, source(i)%extensionFlag) !




  

200 format("fk5;ellipse(",  f10.6, ", ",  f10.6, ", ",         &
              f10.6, ", ",  f10.6, ", ", f6.2, ") # color=", a,  &
              " text={", a, "}")  




100 format("fk5;ELLIPSE(",  f10.6, ", ",  f10.6, ", ",         &
              f10.6, ", ",  f10.6, ", ", f6.2, ") # color=", a)




       
        if ( isnullValue(source(i)%semia) .or. source(i)%semia .lt. 0.0) then
          
            theta = 0.0  
            if (plotNos) then
                write(ioUnit, 200) &
                    source(i)%rightascension, source(i)%declination,   & 
                    minWidth,  &   ! 
                    minWidth,  &   ! 
                    theta, trim(colour), trim(number)
            else
                write(ioUnit, 100) &
                    source(i)%rightascension, source(i)%declination,   & 
                    minWidth,  &   ! 
                    minWidth,  &   ! 
                    theta, trim(colour)
            end if
 

        else
            if (plotNos) then
                write(ioUnit, 200) &
                    source(i)%rightAscension, source(i)%declination,   & 
                    max(minWidth, source(i)%semiA  / 3600.0), &   ! 
                    max(minWidth, source(i)%semiB  / 3600.0), &   ! 
                    source(i)%theta, trim(colour), trim(number)
            else

                write(ioUnit, 100) &
                    source(i)%rightAscension, source(i)%declination,   & 
                    max(minWidth, source(i)%semiA  / 3600.0), &   ! 
                    max(minWidth, source(i)%semiB  / 3600.0), &   ! 
                    source(i)%theta, trim(colour)
      
            end if


        end if



    end do
    close(ioUnit)
    
end subroutine CreateRegionFile11
 
 
!*********************************************
! CreateRegionFile
!*********************************************
subroutine CreateRegionFile(nSources, source, regionFile, plotNumbers)
    integer(int32), intent(in) :: nSources

    type(sourceParameters), pointer :: source(:)

    character(*), intent(in) :: regionFile
    logical, optional :: plotNumbers
    logical :: plotNos  
    integer(int32) :: i, j
    integer(int32), parameter :: ioUnit = 2
    character(10) :: colour
    character(10) :: number  
    real(single) :: minWidth, theta, semia, semib
    real(single), parameter :: dr = 0.5

   





    if ( present(plotNumbers)) then
        plotNos = plotNumbers
    else
        plotNos = .true.
    end if


    open(unit = ioUnit, file = regionfile, status = 'replace')
!    write(ioUnit,"(a)") '# Region file format: DS9 version 3.0'
!    write(ioUnit,"(a)") 'global color=green font="helvetica &
!    &10 normal" edit=0 move=0 delete=1 include=1 fixed=0'

   
    write(iounit, *) "global select=0 edit=0 move=0 delete=1"


    minWidth = 1.0 / 3600.0

    do i = 1, nSources

        write(number, '(i5)') i
        number = adjustl(number)
!        colour = RegionColour(source(i)%confusionFlag, source(i)%qualityFlag, source(i)%extensionFlag) !


        if (source(i)%extensionFlag == 0) then

            colour = "green"
        else
            colour = "cyan"
      
        end if  





!write(*, *) i, " ra=",  source(i)%rightAscension, " dec=",source(i)%declination, &
 ! " a=", source(i)%semia, " b=", source(i)%semib, trim(colour), trim(number)     

200 format("fk5;ellipse(",  f10.6, ", ",  f10.6, ", ",         &
              f10.6, ", ",  f10.6, ", ", f6.2, ") # color=", a,  &
              " text={", a, "}")  




100 format("fk5;ELLIPSE(",  f10.6, ", ",  f10.6, ", ",         &
              f10.6, ", ",  f10.6, ", ", f6.2, ") # color=", a)


        
        if ( isnullValue(source(i)%semia) .or. source(i)%semia .lt. 0.0) then
            semia = minwidth
            semib = minwidth
            theta = 0.0
  
            if (plotNos) then
                 
                write(ioUnit, 200) &
                    source(i)%rightascension, source(i)%declination,   & 
                    semia, semib, theta, trim(colour), trim(number)
            else
                write(ioUnit, 100) &
                    source(i)%rightascension, source(i)%declination,   & 
                    semia, semib, theta, trim(colour)
            end if
 

        else
            semia = max(minwidth, source(i)%semia / 3600.0)
            semib = max(minwidth, source(i)%semib / 3600.0)
            theta = source(i)%theta
            if (plotNos) then
                write(ioUnit, 200) &
                    source(i)%rightAscension, source(i)%declination,   & 
                    semia, semib, theta, trim(colour), trim(number)
            else
                

                write(ioUnit, 100) &
                    source(i)%rightAscension, source(i)%declination,   & 
                    semia, semib, theta, trim(colour)
      
            end if


        end if
        !************************************************************************
        !
        !************************************************************************
        do j = 0, size(colours) - 1
            if (btest(source(i)%qualityFlag, j)) then
             
                                     
                semia = semia + dr / 3600.0
                semib = semib + dr / 3600.0

                write(ioUnit, 100) &
                    source(i)%rightAscension, source(i)%declination,   & 
                    semia, semib,                                      & 
                    theta, trim(colours(j))




            end if
        end do


  

    end do
    close(ioUnit) 

end subroutine CreateRegionFile
 




!*********************************************
! CreateRegionFile
!*********************************************
subroutine CreateRegionFile1(nSources, source, regionFile, plotNumbers)
    integer(int32), intent(in) :: nSources

    type(sourceParameters), pointer :: source(:)

    character(*), intent(in) :: regionFile
    logical, intent(in), optional :: plotNumbers

    integer(int32) :: i, j
    integer(int32), parameter :: ioUnit = 2
 
    character(10) :: colour
    character(10) :: number
   

    real(single) :: semia, semib, theta
    logical :: plotNos

  

 

    if ( present(plotNumbers) ) then
        plotNos = plotNumbers
    else
        plotNos = .true.
    end if

    open(unit = ioUnit, file = regionfile, status = 'replace')
!    write(ioUnit,"(a)") '# Region file format: DS9 version 3.0'
!    write(ioUnit,"(a)") 'global color=green font="helvetica &
!    &10 normal" edit=0 move=0 delete=1 include=1 fixed=0'


    write(iounit, *) "global select=0 edit=0 move=0 delete=1"

    do i = 1, nSources
        write(number, '(i5)') i
        number = adjustl(number)
!        colour = RegionColour(source(i)%confusionFlag, source(i)%qualityFlag, source(i)%extensionFlag) 


     
            if (source(i)%extensionFlag == 0) then

                colour = "green"
            else
                colour = "cyan"
      
            end if  
       




       

100 format("ellipse(",  f10.4, ", ",  f10.4, ", ",         &
              f8.2, ", ",  f8.2, ", ", f6.2, ") # color=", a, &        
              " text={", a, "}")  
  

200 format("ellipse(",  f10.4, ", ",  f10.4, ", ",         &
              f8.2, ", ",  f8.2, ", ", f6.2, ") # color=", a)
  

          
        if ( isnullvalue(source(i)%semia) .or. source(i)%semia .lt. 0.0) then

            semia = 2.0
            semib = 2.0
            theta = 0.0  


            if (plotNos) then
                write(ioUnit, 100) &
                    source(i)%xpos, source(i)%ypos,   & 
                    semiA,  &   ! 
                    semiB, &   ! 
                    theta, trim(colour), trim(number)
            else

                write(ioUnit, 200) &
                    source(i)%xpos, source(i)%ypos,   & 
                    semiA,  &   ! 
                    semiB, &   ! 
                    theta, trim(colour)
            end if



        else

            semia = max(1.0, source(i)%semia * 3.0)
            semib = max(1.0, source(i)%semib * 3.0)
             
            theta = source(i)%theta

            if (plotNos) then

          
                write(ioUnit, 100) &
                      source(i)%xpos, source(i)%ypos,   & 
                      semia, semib, theta, trim(colour), trim(number)

            else

           
                write(ioUnit, 200) &
                      source(i)%xpos, source(i)%ypos,   & 
                      semia, semib, theta, trim(colour)
      
            end if

        end if  

        !************************************************************************
        !
        !************************************************************************
        do j = 0, size(colours) - 1
            if (btest(source(i)%qualityFlag, j)) then
             
                                     
                semia = semia + 0.75
                semib = semib + 0.75

                write(ioUnit, 200) &
                    source(i)%xpos, source(i)%ypos,                    & 
                    semia, semib,                                      & 
                    theta, trim(colours(j))

!                write(*, *) "source ", i, " x=", source(i)%xpos, " y=", source(i)%ypos, &
!                             " quality bit ", j, " set- colour=", trim(colours(j))


            end if

        end do
     

    
    end do
    close(ioUnit)
    
end subroutine CreateRegionFile1





!*********************************************
! CreateRegionFile
!*********************************************
subroutine CreateRegionFile51(nSources, source, regionFile, plotNumbers, colourIn)
    integer(int32), intent(in) :: nSources

    type(sourceParameters), pointer :: source(:)

    character(*), intent(in) :: regionFile
    logical, intent(in), optional :: plotNumbers

    integer(int32) :: i, j
    integer(int32), parameter :: ioUnit = 2
    character(*), intent(in), optional :: colourIn
    character(10) :: colour
    character(10) :: number
    logical :: noColour = .true.

    real(single) :: semia, semib, theta
    logical :: plotNos

  

    if(present(colourIn)) then

        colour = trim(colourIn)
        noColour = .false.
    else
        noColour = .true.
    end if

    if ( present(plotNumbers) ) then
        plotNos = plotNumbers
    else
        plotNos = .true.
    end if

    open(unit = ioUnit, file = regionfile, status = 'replace')
!    write(ioUnit,"(a)") '# Region file format: DS9 version 3.0'
!    write(ioUnit,"(a)") 'global color=green font="helvetica &
!    &10 normal" edit=0 move=0 delete=1 include=1 fixed=0'


    write(iounit, *) "global select=0 edit=0 move=0 delete=1"

    do i = 1, nSources

        write(number, '(i5)') i
        number = adjustl(number)
!        colour = RegionColour(source(i)%confusionFlag, source(i)%qualityFlag, source(i)%extensionFlag) 


        if(noColour) then
            if (source(i)%extensionFlag == 0) then

                colour = "green"
            else
                colour = "cyan"
      
            end if  
        end if




       

100 format("ellipse(",  f10.4, ", ",  f10.4, ", ",         &
              f8.2, ", ",  f8.2, ", ", f6.2, ") # color=", a, &        
              " text={", a, "}")  
  

200 format("ellipse(",  f10.4, ", ",  f10.4, ", ",         &
              f8.2, ", ",  f8.2, ", ", f6.2, ") # color=", a)
  

          
        if ( isnullvalue(source(i)%semia) .or. source(i)%semia .lt. 0.0) then

            semia = 2.0
            semib = 2.0
            theta = 0.0  


            if (plotNos) then
                write(ioUnit, 100) &
                    source(i)%xpos, source(i)%ypos,   & 
                    semiA,  &   ! 
                    semiB, &   ! 
                    theta, trim(colour), trim(number)
            else

                write(ioUnit, 200) &
                    source(i)%xpos, source(i)%ypos,   & 
                    semiA,  &   ! 
                    semiB, &   ! 
                    theta, trim(colour)
            end if



        else

            semia = max(1.0, source(i)%semia * 3.0)
            semib = max(1.0, source(i)%semib * 3.0)
             
            theta = source(i)%theta

            if (plotNos) then

          
                write(ioUnit, 100) &
                      source(i)%xpos, source(i)%ypos,   & 
                      semia, semib, theta, trim(colour), trim(number)

            else

           
                write(ioUnit, 200) &
                      source(i)%xpos, source(i)%ypos,   & 
                      semia, semib, theta, trim(colour)
      
            end if

        end if  

        !************************************************************************
        !
        !************************************************************************
        do j = 0, size(colours) - 1
            if (btest(source(i)%qualityFlag, j)) then
             
                                     
                semia = semia + 0.75
                semib = semib + 0.75

                write(ioUnit, 200) &
                    source(i)%xpos, source(i)%ypos,                    & 
                    semia, semib,                                      & 
                    theta, trim(colours(j))

!                write(*, *) "source ", i, " x=", source(i)%xpos, " y=", source(i)%ypos, &
!                             " quality bit ", j, " set- colour=", trim(colours(j))


            end if

        end do
     

    
    end do
    close(ioUnit)
    
end subroutine CreateRegionFile51








!*********************************************
! CreateRegionFile21
!*********************************************
subroutine CreateRegionFile21(nSources, source, regionFile, colour)
    integer(int32), intent(in) :: nSources

    type(sourceParameters), pointer :: source(:)

    character(*), intent(in) :: regionFile
    character(*) :: colour

    integer(int32) :: i
    integer(int32), parameter :: ioUnit = 2
   
    character(10) :: number
    real(single) :: semia, semib, theta
   

  

  
    open(unit = ioUnit, file = regionfile, status = 'replace')



    write(iounit, *) "global select=0 edit=0 move=0 delete=1"
    do i = 1, nSources

        write(number, '(i5)') i
        number = adjustl(number)

       

100 format("ellipse(",  f10.4, ", ",  f10.4, ", ",         &
              f8.2, ", ",  f8.2, ", ", f6.2, ") # color=", a, &        
              " text={", a, "}")  
  

200 format("ellipse(",  f10.4, ", ",  f10.4, ", ",         &
              f8.2, ", ",  f8.2, ", ", f6.2, ") # color=", a)
  

          
        if ( isnullvalue(source(i)%semia) .or. source(i)%semia .lt. 0.0) then

            semia = 2.0
            semib = 2.0
            theta = 0.0  


           
            write(ioUnit, 100) &
                    source(i)%xpos, source(i)%ypos,   & 
                    semiA,  &   ! 
                    semiB, &   ! 
                    theta, trim(colour), trim(number)
           



        else

            semia = max(1.0, source(i)%semia * 3.0)
            semib = max(1.0, source(i)%semib * 3.0)
             
            theta = source(i)%theta

           

          
            write(ioUnit, 100) &
                      source(i)%xpos, source(i)%ypos,   & 
                      semia, semib, theta, trim(colour), trim(number)

        

        end if  

  
    
    end do
    close(ioUnit)
    
end subroutine CreateRegionFile21











!*********************************************
! CreateRegionFile
!*********************************************
subroutine CreateRegionFile100(nSources, source, regionFile, plotNumbers)
    integer(int32), intent(in) :: nSources

    type(sourceParameters), pointer :: source(:)

    character(*), intent(in) :: regionFile
    logical, intent(in), optional :: plotNumbers

    integer(int32) :: i
    integer(int32), parameter :: ioUnit = 2
    character(10) :: colour
    character(10) :: number
    real(single) :: semia, semib, theta
    logical :: plotNos


    if ( present(plotNumbers) ) then
        plotNos = plotNumbers
    else
        plotNos = .true.
    end if





  
    open(unit = ioUnit, file = regionfile, status = 'replace')
!    write(ioUnit,"(a)") '# Region file format: DS9 version 3.0'
!    write(ioUnit,"(a)") 'global color=green font="helvetica &
!    &10 normal" edit=0 move=0 delete=1 include=1 fixed=0'


    write(iounit, *) "global select=0 edit=0 move=0 delete=1"
    do i = 1, nSources

        write(number, '(i5)') i
        number = adjustl(number)
        colour = RegionColour(source(i)%confusionFlag, source(i)%qualityFlag, source(i)%extensionFlag) 


       

100 format("ellipse(",  f10.4, ", ",  f10.4, ", ",         &
              f8.2, ", ",  f8.2, ", ", f6.2, ") # color=", a, &        
              " text={", a, "}")  
  

200 format("ellipse(",  f10.4, ", ",  f10.4, ", ",         &
              f8.2, ", ",  f8.2, ", ", f6.2, ") # color=", a)
  

          
        if ( isnullvalue(source(i)%semia) .or. source(i)%semia .lt. 0.0) then

            semia = 2.0
            semib = 2.0
            theta = 0.0  


            if (plotNos) then
                write(ioUnit, 100) &
                    source(i)%xpos, source(i)%ypos,   & 
                    semiA,  &   ! 
                    semiB, &   ! 
                    theta, trim(colour), trim(number)
            else

                write(ioUnit, 200) &
                    source(i)%xpos, source(i)%ypos,   & 
                    semiA,  &   ! 
                    semiB, &   ! 
                    theta, trim(colour)
            end if



        else
            if (plotNos) then
!source(i)%theta =0.0
                write(ioUnit, 100) &
                      source(i)%xpos, source(i)%ypos,   & 
!                      max(1.0, source(i)%semiA),  &   ! 
!                      max(1.0, source(i)%semiB), &   ! 
              


                      max(1.0, source(i)%semiA * 3.0),  &   ! 
                      max(1.0, source(i)%semiB * 3.0), &   ! 
                      source(i)%theta, trim(colour), trim(number)



            else
 
                write(ioUnit, 200) &
                      source(i)%xpos, source(i)%ypos,   & 
                      max(1.0, source(i)%semiA * 3.0),  &   ! 
                      max(1.0, source(i)%semiB * 3.0), &   ! 
                      source(i)%theta, trim(colour)
      
            end if

        end if  
    end do
    close(ioUnit)
    
end subroutine CreateRegionFile100





!**************************************************
!
!*************************************************
character(10) Function RegionColour(cflag, qFlag, eFlag)
    integer(int8), intent(in) :: cflag, eFlag
    integer(int16), intent(in) :: qflag
    integer(int16) :: i

    if (qFlag .gt. 0) then
        do i = 0, size(colours) - 1
            if( btest(qFlag, i)) then
                RegionColour = colours(i)
                return
            end if
        end do
      


    elseif (btest(cflag, 2)) then
        regionColour = "blue"
      
    elseif (btest(cflag, 4)) then
        regionColour = "red"
   
  

    else if (eFlag .eq. 0) then
        regionColour = "green"
    else
        regionColour = "cyan"
      
    end if  
!write(*, *) "qflag=", qflag, " colour=", regionColour

end function RegionColour
!*******************************************************************
!
!******************************************************************
subroutine GetWindowParameters(sourceListDataSetH, windowX0, windowY0, xbin, ybin)
    
    type(dataSetT) :: sourceListDataSetH

    integer(int32), intent(out) :: windowX0, windowY0, xbin, ybin
    logical :: binBpe
 
 
    character(100) :: windowString
    character(1) :: c
    integer(int32) :: windowX1, windowY1, winDx, winDy

    xbin = int32Attribute(sourceListDataSetH, "BINAX1")      ! binning factors
    ybin = int32Attribute(sourceListDataSetH, "BINAX2")
    binbpe = booleanattribute(sourceListDataSeth, "BINBPE")


    if (xbin .ne. ybin) then
        call fatal("diffBin", "This version cannot deal with different binning factors in x and y")
    end if

    if (hasAttribute(sourceListDataSetH, "DETSEC")) then
        windowString = stringAttribute(sourceListDataSetH, "DETSEC")
        read (windowString, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
                  c, windowX0, c, windowX1, c,  windowY0, c, windowY1, c

        winDX = windowX1 - windowX0 + 1
        winDY = windowY1 - windowY0 + 1


    end if

    binBPE = booleanAttribute(sourceListDataSetH, "BINBPE")

    ! Calculate the image binning.
    xBin = 2 ** xbin
    yBin = 2 ** ybin

    if (binBPE) then
        xBin = xBin * 2
        yBin = yBin * 2
    end if

    windowX0 = windowX0 * xBin
    windowY0 = windowY0 * yBin

end subroutine GetWindowParameters









!********************************************************************
!
!*******************************************************************
logical Function isNullValue1(x)

    real(single), intent(in) :: x
    if (.not. (x > 0.0) .and. .not. (x <= 0.0)) then
        isNullValue1 = .true.
    else
        isNullValue1 = .false.
    end if

end function IsNullValue1

!********************************************************************
!
!*******************************************************************
logical Function isNullValue2(x)

    real(double), intent(in) :: x
    if (.not. (x > 0.0) .and. .not. (x <= 0.0)) then
        isNullValue2 = .true.
    else
        isNullValue2 = .false.
    end if

end function IsNullValue2






end module OmUtility_mod

















