! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module plot_region_mod

use dal

implicit none

integer, parameter :: MAX_PTS = 50

contains

subroutine plot_fits_region(filename, ci)

implicit none
character(len=*) :: filename
character(len=20) :: shape
integer i,j,k,n,ci
integer :: nRows
real(kind=Single), pointer :: yval(:,:), xval(:,:), rval(:,:)
type(DataSetT) :: regDataSet
type(TableT) :: tableSet
type(ColumnT) :: shapeCol, xCol, yCol, rCol
call pgqci(k)
call pgsci(ci)

write(*,*) k

regDataSet = dataSet(filename, read)
tableSet = table(regDataSet, "REGION")
nRows = numberOfRows(tableSet)
shapeCol = column(tableSet,"SHAPE", read)


do i = 0, nRows-1

  shape = stringCell(shapeCol, i)

  if ((shape.eq."POLYGON").or.(shape.eq."!POLYGON")) then
    xCol = column(tableSet,"X",read)
    xval => real32VectorData(xCol)
    yCol = column(tableSet,"Y",read)
    yval => real32VectorData(yCol)
    do j = 1, MAX_PTS-1
       if ((xval(j,i) == xval(0,i)).and. &
           (yval(j,i) == yval(0,i)) ) then
             n = j
             EXIT 
       endif
    enddo
       
   call pgmove(xval(0,i),yval(0,i))
   do j = 1, n-1
    call pgdraw(xval(j,i),yval(j,i))
   enddo
   call pgdraw(xval(0,i),yval(0,i))
   call release(xCol)
   call release(yCol)
  endif

  if ((shape.eq."CIRCLE").or.(shape.eq."!CIRCLE")) then
    xCol = column(tableSet,"X",read)
    xval => real32VectorData(xCol)
    yCol = column(tableSet,"Y",read)
    yval => real32VectorData(yCol)
    rCol = column(tableSet,"R",read)
    rval => real32VectorData(rCol)
    call pgcirc(xval(0,i), yval(0,i), rval(0,i))
    call release(xCol)
    call release(yCol)
    call release(rCol)
  endif


  if ((shape.eq."ANNULUS").or.(shape.eq."!ANNULUS")) then
    xCol = column(tableSet,"X",read)
    xval => real32VectorData(xCol)
    yCol = column(tableSet,"Y",read)
    yval => real32VectorData(yCol)
    rCol = column(tableSet,"R",read)
    rval => real32VectorData(rCol)
    call pgcirc(xval(0,i), yval(0,i), rval(0,i))
    call pgcirc(xval(0,i), yval(0,i), rval(1,i))
    call release(xCol)
    call release(yCol)
    call release(rCol)
  endif

enddo
call release(shapeCol)
call release(regDataSet)



call pgsci(k)
end subroutine plot_fits_region

      SUBROUTINE SSTRIP(STRING)
!
! Subroutine to strip leading spaces from a character string
!
      CHARACTER STRING*(*)
      INTEGER I
      IF ( STRING.NE.' ' ) THEN
        I = 1
10000   IF ( STRING(I:I).EQ.' ' ) THEN
          I = I+1
          GOTO 10000
        END IF
        STRING = STRING(I:)
      END IF
      END SUBROUTINE SSTRIP

end module plot_region_mod
