! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module readHist_mod

use dal

implicit none

private

public readHist

contains

subroutine readHist(histFile, x, y, roll, nhist, nbad)


character*(*) histFile
integer nhist,nbad
real, pointer :: x(:),y(:),roll(:)
type(DataSetT) :: historyDataSet
type(TableT) :: tableDataSet
type(ColumnT) :: pcolumn
integer(kind=int32), pointer :: idx(:),idy(:),iroll(:),iquality(:),nggs(:)
real, allocatable :: quality (:)
integer i,nrows
real, parameter :: TBD = 10.

historyDataSet = dataSet(histFile, read)

tableDataSet = table(historyDataSet,0)
nrows = numberOfRows(tableDataSet)

pcolumn = column(tableDataSet, "DX", read)
idx => int32data(pcolumn)
pcolumn = column(tableDataSet, "DY", read)
idy => int32data(pcolumn) 
pcolumn = column(tableDataSet, "ROLL", read)
iroll => int32data(pcolumn)
pcolumn = column(tableDataSet, "QUALITY", read)
iquality => int32data(pcolumn)
pcolumn = column(tableDataSet, "NGGS", read)
nggs => int32data(pcolumn)

allocate(quality(0:nrows-1))

do i = 0 , nrows-1
   quality(i) = real(iquality(i))/1000000.
   quality(i) = sqrt(quality(i) / (2.*real(nggs(i))-1.))
enddo

allocate(x(0:nrows-1))
allocate(y(0:nrows-1))
allocate(roll(0:nrows-1))

nbad = 0
nhist = 0
do i = 0 , nrows-1
 if (quality(i) < TBD) then             ! what constitutes a bad tracking soln?
   x(nhist) = real (idx(i))/1000.
   y(nhist) = real (idy(i))/1000.
   roll(nhist) = asin(real (iroll(i))/1000000.)
   nhist = nhist + 1
 else
  nbad = nbad + 1
 endif
enddo

   
call release(tableDataSet)
call release(historyDataSet)
end subroutine readHist

end module readHist_mod
